# List of functions to correct for errors in data
import numpy as np
import pandas as pd
import re

def isBadSession(subj,task,sess):
    badsessions = pd.read_pickle('/home1/daniel.rubinstein/Projects/dataQC/badsessions.pkl')
    check = badsessions[(badsessions['subject']==subj) & (badsessions['task']==task) & (badsessions['session']==int(sess))]
    isbad = (len(check) > 0)
    
    return isbad


def correctEEGoffset(sub,session,exp,reader,events):
    # The EEG for many FR subjects (FR1 and catFR1 in particular) does not align with the events since the 
    # implementation of Unity. This is a temporary fix for the EEG alignment for these subjects before
    # the data is corrected in Rhino. Subject-by-subject details are here:
    # https://docs.google.com/spreadsheets/d/1co5f7-dPOktGIXZJ7uptv0SwBJhf36TuhVSMFqRC0X8/edit?usp=sharing
    # JS 2020-09-22
    # Update 2020-09-29 accounting for sampling rate and raising error if eeg events don't exist
    
    ## Inputs ##
    # sub: subject name (type str)
    # session: session number (type int)
    # exp: experiment, typically 'FR1' or 'catFR1' (type str)
    # reader: typical output from CMLReader function (see cmlreaders documentation)
    # events: dataFrame from reader.load('task_events') for your events of choice
    
    sub_num = [int(s) for s in re.findall(r'\d+',sub)] # extract number for sub   
    
    temp_eeg = reader.load_eeg(events=events, rel_start=0, rel_stop=100) # just to get sampling rate
    sr = temp_eeg.samplerate
    sr_factor = 1000/sr
    
    if sum(events.eegoffset==-1)>0:
        raise Exception('Events without EEG (those with events.eegoffset=-1) should be removed before calling correctEEGoffset')
 
    if (sub in ['R1379E','R1385E','R1387E','R1394E','R1402E']) or \
        (sub=='R1404E' and session==0 and exp=='catFR1'): 
        # first 5 true for catFR1 and FR1. R1404E only one catFR1 session has partial beep 
        # for these subs there is a partial beep and 500 ms of eeg lag (see "History of issues 2020-09-08" for examples)
        
        # add time (in units of samples) since the events are already ahead of the eeg
        events.eegoffset = events.eegoffset+int(np.round(500/sr_factor))
        
    # subs where unity was implemented for some sessions but not others
    elif (sub=='R1396T' and exp=='catFR1') or (sub=='R1396T' and session==1) or \
         (sub=='R1395M' and exp=='catFR1') or (sub=='R1395M' and exp=='FR1' and session>0):
        events.eegoffset = events.eegoffset+int(np.round(1000/sr_factor))
        
    # do nothing since these sessions were pyEPL so the offset is okay
    elif (sub=='R1406M' and session==0) or (sub=='R1415T' and session==0 and exp=='FR1') or (sub=='R1422T' and exp=='FR1'):
        pass 
    
    # remaining unity subs
    elif sub_num[0]>=1397 or sub == 'R1389J': 
        events.eegoffset = events.eegoffset+int(np.round(1000/sr_factor))
        
    return events


def fixMNEoffset(eeg,eeg_mne,reader,evs,start,stop,elecs):
    import numpy as np
    offsets = [0,-1,1,-2,2]
    i=0 #initialize
    eeg_i = eeg
    eeg_mne_i = eeg_mne
    while (eeg.data.shape[-1] != len(eeg_mne.times)) & (i < len(offsets)-1):
        i+=1
        offset = offsets[i]
        eeg = reader.load_eeg(events=evs, rel_start=start+offset, rel_stop=stop, scheme=elecs)
        eeg_mne = eeg.to_mne()
    if eeg.data.shape[-1] != len(eeg_mne.times):
        return eeg_i, eeg_mne_i, np.nan
    else:
        return eeg, eeg_mne, offsets[i]

    
def checkBipDist(reader,pairs,thresh=20):
    from cmlreaders import CMLReader
    from scipy.spatial import distance
    
    # Function to check the distance between the individual elecs in bipolar pairs
    # Returns boolean array of True or False
    if isinstance(pairs,pd.DataFrame):
        results = np.ones(len(pairs),dtype='bool')
    elif isinstance(pairs,pd.Series):
        results = True
    
    # Try to load monopolar contacts
    try:
        monos = reader.load("contacts")
    except:
        print('cannot load contacts for subject',reader.subject)
        return results
    
    if 'avg.corrected.x' in pairs:
        xcoord, ycoord, zcoord = 'avg.corrected.x','avg.corrected.y','avg.corrected.z'
    elif 'avg.dural.x' in pairs:
        xcoord, ycoord, zcoord = 'avg.dural.x','avg.dural.y','avg.dural.z'
    else:
        xcoord, ycoord, zcoord = 'avg.x','avg.y','avg.z'
            
    # Calculate the distance between each bipolar pair contact
    if isinstance(pairs,pd.DataFrame):
        bip_dists = []
        for p, pair in pairs.iterrows():
            coord_a = monos[monos.contact==pair['contact_1']][[xcoord, ycoord, zcoord]]
            coord_b = monos[monos.contact==pair['contact_2']][[xcoord, ycoord, zcoord]]
            try:
                bip_dist = distance.euclidean(coord_a,coord_b)
            except:
                bip_dist = np.nan

            bip_dists.append(bip_dist)
            results[p] = bip_dist < thresh
    elif isinstance(pairs,pd.Series):
        coord_a = monos[monos.contact==pairs['contact_1']][[xcoord, ycoord, zcoord]]
        coord_b = monos[monos.contact==pairs['contact_2']][[xcoord, ycoord, zcoord]]
        try:
            bip_dists = distance.euclidean(coord_a,coord_b)
        except:
            bip_dists = np.nan

        results = bip_dist < thresh
        
    return results, bip_dists


def removeOutliers(powsa,ax=3):
    # Assumes structure of event x elec x freq x time
    # Remove outliers from a log power array, based on variance across one axis and count the number of elecs
    maxval = np.mean(powsa,ax) + 5*np.std(powsa,ax)
    maxval = np.expand_dims(maxval,ax)
    maxval = np.repeat(maxval,powsa.shape[3],axis=ax)

    # If >20% of timepoints exceed threshold (in any one frequency for a given elec), remove that list for that elec
    pct = np.zeros(powsa.shape[0:3])
    badelecs = []
    for ev in range(powsa.shape[0]):
        for elec in range(powsa.shape[1]):
            for f in range(powsa.shape[2]):
                p = powsa[ev,elec,f]
                mv = maxval[ev,elec,f]
                pct[ev,elec,f] = np.sum([p > mv])/np.product(p.shape)
                if pct[ev,elec,f] > 0.2:
                    powsa[ev,elec] = np.nan
                    badelecs.append((ev,elec))
                    continue
    
    return powsa, badelecs
 