# Module for functions for list SME project

from cmlreaders import CMLReader, get_data_index
import numpy as np
import pandas as pd
from mne import time_frequency, filter, EpochsArray
from scipy.stats import pearsonr, spearmanr, zscore
from statsmodels.stats.multitest import multipletests 
from localization import getRegions, checkBadElectrodes
from eegQC import fixMNEoffset, checkBipDist

def justfinites(arr):
    return np.array(arr)[np.isfinite(arr)]


def vecshift(vec,shift):
    dim = len(vec)
    newidx = np.mod(np.arange(dim)+shift,dim)
    newvec = [vec[a] for a in newidx]

    return np.array(newvec)


def shuffleItemRecalls(result,norm=False):
    '''Function to shuffle the recalled items in a dataframe, within session''' 

    # For each session shuffle the recalls and make sure to update corresponding list perfs array
    for r, recalls in enumerate(result['recalls']):
        # First create shuffle index
        shuffle = np.random.permutation(len(recalls['recalled']))
        # Then get list of recalls and shuffle
        recalled = np.array(recalls['recalled'])
        recalls['recalled'] = list(recalled[shuffle])
        if norm:
            recalled_norm = np.array(recalls['recalled_norm'])
            recalled_norm_holdout = np.array(recalls['recalled_norm_holdout'])
            recalls['recalled_norm'] = list(recalled_norm[shuffle])
            recalls['recalled_norm_holdout'] = list(recalled_norm_holdout[shuffle])
        
        # update original dataframe
        result['recalls'][r] = recalls
        
        # update list perfs
        result['perfs_list'][r] = list(recalls.groupby('list').sum()['recalled'])

    return result


def getListPerfs(subj, task, sps=[1,2,3,4,5,6,7,8,9,10,11,12]):
    from eegQC import isBadSession
    
    df = get_data_index("r1")
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])

    listperfs = []
    for s, sess in enumerate(sessions):
        if isBadSession(subj,task,sess):
            continue
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s])
        evs = reader.load('task_events')
        evs = evs[evs['eegoffset'] > -1]
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        
        # Only use lists of length 12 and that have full recall periods
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete_sess = word_evs['list'].unique()[[n==12 for n in nwords]]
        lists_withrec = evs[evs['type']=='REC_START']['list'].unique()
        lists_withend = evs[evs['type']=='REC_END']['list'].unique()
        lists_complete_sess = list(set(lists_withrec).intersection(set(lists_complete_sess)))
        if subj == 'R1342M':
            lists_complete_sess = list(set(lists_withend).intersection(set(lists_complete_sess)))
        if ((subj == 'R1395M') & (task == 'catFR1') & (sess == 1)):
            lists_complete_sess.remove(10)
        word_evs = word_evs[[l in lists_complete_sess for l in word_evs['list']]]
        word_evs = word_evs[[sp in sps for sp in np.array(word_evs['serialpos'])]] #restrict to certain serial positions
        
        recalls = list(word_evs[word_evs['recalled']==True]['list'])
        listperfs.append([sum(recalls==l) for l in word_evs['list'].unique()])
    
    return listperfs


def getProbLeadRecall(subj,task,lead,chain='backward'):
    from eegQC import isBadSession
    # At a session level, get the probability of recalling serial position i+[lead], given recall of i
    prob_leadrecall = []
    df = get_data_index("r1")
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])

    for s, sess in enumerate(sessions):
        yays=0
        nays=0
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s])   #reader for loading CML data
        if isBadSession(subj,task,sess):
            continue
        evs = reader.load('task_events')
        evs = evs[evs['eegoffset'] > -1]
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        
        # Only use lists of length 12 and also has a recall period
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete_sess = word_evs['list'].unique()[[n==12 for n in nwords]]
        lists_withrec = evs[evs['type']=='REC_START']['list'].unique()
        lists_withend = evs[evs['type']=='REC_END']['list'].unique()
        lists_complete_sess = list(set(lists_withrec).intersection(set(lists_complete_sess)))
        if subj == 'R1342M':
            lists_complete_sess = list(set(lists_withend).intersection(set(lists_complete_sess)))
        if ((subj == 'R1395M') & (task == 'catFR1') & (sess == 1)):
            lists_complete_sess.remove(10)
        word_evs = word_evs[[l in lists_complete_sess for l in word_evs['list']]]

        for ll in lists_complete_sess:
            listwords = word_evs[word_evs['list']==ll]
            recalledwords = listwords[listwords['recalled']==True]
            listrecalls_sp = np.array(recalledwords['serialpos'])
            # Method 1, forward chaining
            if chain=='forward':
                # Get list recalls by serial position, ignore if serial position is more than 12-lead
                for listrecall_sp in listrecalls_sp[listrecalls_sp <= 12-lead]:
                    # count the number of times recall of sp(i) predicts recall of sp(i+1)
                    if listrecall_sp+lead in listrecalls_sp:
                        yays+=1
                    else:
                        nays+=1
            # or for method 2, backward chaining
            elif chain=='backward':
                for listrecall_sp in listrecalls_sp[listrecalls_sp >= 1+lead]:
                    # count the number of times recall of sp(i) is predicted by recall of sp(i-1)
                    if listrecall_sp-lead in listrecalls_sp:
                        yays+=1
                    else:
                        nays+=1
        # Calculate the probability of lead recall
        prob_leadrecall.append(yays/(yays+nays))

    return prob_leadrecall


def getListPowCorr0(subj, task, myfreqs, rois='all', tois=[0,30000], tint=3000):
    #The "r1" database corresponds to all of the RAM subjects
    df = get_data_index("r1")

    #sub = 'R1111M'
    #task = 'FR1'
    #tois = [0,30000]
    #tint = 3000
    times = np.linspace(tois[0],tois[1]-tint,int(tois[1]/tint))+tint/2

    #Find out the sessions, localization, and montage for this subject
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])

    rs_z = []
    ps = []
    ps_fdr = []
    perfs_list = []
    perfs_z = []
    pows_z = []
    
    for s, sess in enumerate(sessions):
        # For first session...
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s])   #reader for loading CML data
        evs = reader.load('task_events')
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        recalls = word_evs[['recalled','list']]
        list_evs = evs[(evs['type']=='TRIAL') & (evs['list']>0)]
        words_first = word_evs[word_evs['serialpos']==1]
        
        # Only use the complete encoding (length 12) lists
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete = word_evs['list'].unique()[[n==12 for n in nwords]]
        word_evs = word_evs[[l in lists_complete for l in word_evs['list']]]
        list_evs = list_evs[[l in lists_complete for l in list_evs['list']]]
        words_first = words_first[[l in lists_complete for l in words_first['list']]]

        # Get the pairs
        pairs = reader.load('pairs')

        # Narrow down the pairs
        if rois != 'all':
            regionStr = 'ind.corrected.region'
            if 'ind.corrected.region' not in pairs.columns:
                if 'ind.dural.region' in pairs.columns:
                    regionStr = 'ind.dural.region'
                else:
                    regionStr = 'ind.region'
            regions = np.array(pairs[regionStr])
            pairs = pairs[[r in rois for r in regions]]
        
        # Get the list performances for complete lists
        perfs_list_sess = [np.sum(recalls[recalls['list']==l]['recalled']) for l in lists_complete]
        perfs_list.append(perfs_list_sess)
        perfs_z.append(zscore(perfs_list_sess))
        # Get the contacts in frontal cortex
        #pairs_frontal, pairs_temporal, pairs_parietal = getRegions(reader, pairs)

        # Get the eeg for all lists
        buf = 1000
        eeg = reader.load_eeg(events=words_first, rel_start=tois[0]-buf, rel_stop=tois[1]+buf, scheme=pairs)
        eeg_base = reader.load_eeg(events=list_evs, rel_start=0-buf, rel_stop=3000+buf, scheme=pairs)
        eeg = eeg.to_mne()
        eeg_base = eeg_base.to_mne()

        sr = eeg.info['sfreq']
        
        # Get the band-passed power
        #Pre-define an array of frequencies. We can log-space them since wavelets are less sensitive to differences at higher frequencies
        #myfreqs = np.logspace(np.log10(3), np.log10(150), num=15)
        #myfreqs = [3,4,5,6,7,8]

        #Run TF decomposition and log
        power = time_frequency.tfr_morlet(eeg, freqs=myfreqs, n_cycles=6, return_itc=False, average=False)
        power_base = time_frequency.tfr_morlet(eeg_base, freqs=myfreqs, n_cycles=6, return_itc=False, average=False)
        powsa = np.log10(power.data)
        powsb = np.log10(power_base.data)
        
        # Remove buffer
        buf_samps = int(round(sr*buf/1000.))
        powsa = powsa[:,:,:,buf_samps:-buf_samps] # event x elec x freq x time
        powsb = powsb[:,:,:,buf_samps:-buf_samps]

        # Cut powsa into 10 intervals and average each 3s interval
        nints = int(powsa.shape[-1]/powsb.shape[-1])
        powsa_reshape = np.reshape(powsa,[powsa.shape[0],powsa.shape[1],powsa.shape[2],nints,powsb.shape[-1]])
        powsa_avg = np.nanmean(powsa_reshape,axis=4)

        # Average the power across frequencies
        powsa_avg = np.nanmean(powsa_avg,axis=2)
        
        # Average across frequency and time
        powsb_avg = np.nanmean(powsb,axis=(2,3))
        powsb_avg = np.expand_dims(powsb_avg,axis=2)
        
        # Remove baseline from TOI
        powsc_avg = np.subtract(powsa_avg,powsb_avg)
        
        # Z-score the power
        pows_z_sess = zscore(powsc_avg,axis=0)
        pows_z.append(pows_z_sess)

        # Get the correlations of z-scored power with z-scored list performance
        rs_sess = []
        ps_sess = []

        for e in range(pows_z_sess.shape[1]):
            rs_t = []
            ps_t = []
            for t in range(pows_z_sess.shape[2]):
                r, p = spearmanr(pows_z_sess[:,e,t],zscore(perfs_list_sess))
                rs_t.append(r)
                ps_t.append(p)
            rs_sess.append(rs_t)
            ps_sess.append(ps_t)
        
        rs_sess = np.array(rs_sess)
        ps_sess = np.array(ps_sess)

        ps_fdr_sess = (multipletests(ps_sess.flatten(),method='fdr_bh'))[1]

        # Z-transform the correlations
        rs_z_sess = np.arctanh(rs_sess)
        
        # Append to subject level results
        rs_z.append(rs_z_sess)
        ps.append(ps_sess)
        ps_fdr.append(ps_fdr_sess)
        
        # Append session-level result to subject-level result
        #result = 

    # Aggregate results
    result = {'rs_z':rs_z, 'pvals':ps, 'pvals_fdr':ps_fdr, 'pairs':pairs, 'pows_z':pows_z, 'perfs_list':perfs_list, 'perfs_z':perfs_z, 'myfreqs':myfreqs, 'rois':rois, 'tois':tois, 'times':times}
        
        
    return result
    

def getItemPows(subj, task, myfreqs, rois='all', tois=[300,1600]):
    from scipy.signal import resample
    from eegQC import checkBipDist, removeOutliers
    from localization import get_elec_regions, checkBadElectrodes
    
    df = get_data_index("r1")
    #Find out the sessions, localization, and montage for this subject
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])
    
    perfs_list = []
    itemPows = []
    listPows = []
    recalls = []
    lists_complete = []
    regions = []
    badelecs = []
    
    for s, sess in enumerate(sessions):
        # For first session...
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s])   #reader for loading CML data
        evs = reader.load('task_events')
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        list_evs = evs[(evs['type']=='TRIAL') & (evs['list']>0)]
        words_first = word_evs[word_evs['serialpos']==1]
        
        # Only use lists of length 12
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete_sess = word_evs['list'].unique()[[n==12 for n in nwords]]
        word_evs = word_evs[[l in lists_complete_sess for l in word_evs['list']]]
        list_evs = list_evs[[l in lists_complete_sess for l in list_evs['list']]]
        words_first = words_first[[l in lists_complete_sess for l in words_first['list']]]
        words_lists = list(word_evs['list'])
        
        # Get the list performances for complete lists
        recalls_sess = word_evs[['item_name','recalled','list','serialpos']]
        perfs_list_sess = [np.sum(recalls_sess[recalls_sess['list']==l]['recalled']) for l in lists_complete_sess]
        perfs_list.append(perfs_list_sess)
        
        # Get the pairs
        pairs = reader.load('pairs')
        
        # Fix for abnormally distant electrodes
        valids, _ = checkBipDist(reader,pairs,20)
        pairs = pairs[valids]
        
        # Narrow down the pairs further by checking for bad electrodes
        checks = checkBadElectrodes(reader, pairs)
        pairs = pairs[checks]

        # Get localization
        locz = []
        try:
            locz = reader.load('localization')
        except:
            print('No localization file found')
        regions_sess, _,_,_ = get_elec_regions(locz,pairs)
        
        if rois != 'all':
            regionStr = 'ind.corrected.region'
            if 'ind.corrected.region' not in pairs.columns:
                if 'ind.dural.region' in pairs.columns:
                    regionStr = 'ind.dural.region'
                else:
                    regionStr = 'ind.region'
            #regions_sess = np.array(pairs[regionStr])
            pairs = pairs[[r in rois for r in regions_sess]]
            
        # Get the eeg for all words
        badelecs_sess = []
        
        buf = 1200
        eeg = reader.load_eeg(events=word_evs, rel_start=tois[0]-buf, rel_stop=tois[1]+buf, scheme=pairs)
        eeg = eeg.to_mne()
        sr = eeg.info['sfreq']
        
        #Run TF decomposition and log
        power = time_frequency.tfr_morlet(eeg, freqs=myfreqs, n_cycles=5, return_itc=False, average=False)
        powsa = np.log10(power.data)
        del eeg
        del power
        # Remove buffer
        buf_samps = int(round(sr*buf/1000.))
        powsa = powsa[:,:,:,buf_samps:-buf_samps] # event x elec x freq x time

        # To resample to 50Hz
        #resrate = 50
        #resamps = int((tois[1]-tois[0])/1000*sr/resrate)
        #powsa_res = resample(powsa,resamps,axis=3)
        
        # Remove outlier data points (6 stdevs above mean) by converting to nan
        powsa, badelecs_sess = removeOutliers(powsa,ax=3)
        badelecs.append(badelecs_sess)
        
        # Average power across frequencies
        #powsa_avg = np.nanmean(powsa,axis=2)
        
        # average power within each list, over all lists to get list x elec x freq
        listpows_avg_sess = np.array([np.nanmean(powsa[words_lists == l],axis=(0,3)) for l in np.unique(words_lists)])
        # Just average over time, to get word x elec x freq
        listpows_z_sess = zscore(listpows_avg_sess)
        itempows_avg_sess = np.nanmean(powsa,axis=3)
        del powsa
        itempows_z_sess = zscore(itempows_avg_sess)
        del itempows_avg_sess
        del listpows_avg_sess
        listPows.append(listpows_z_sess)
        itemPows.append(itempows_z_sess)
        recalls.append(recalls_sess)
        lists_complete.append(lists_complete_sess)
        regions.append(regions_sess)
        
    result = {'pairs':pairs, 'itemPows':itemPows, 'lists':lists_complete, 'listPows':listPows, 'recalls':recalls, 'perfs_list':perfs_list, 'pairs_regions':regions, 'myfreqs':myfreqs, 'rois':rois, 'tois':tois, 'badelecs':badelecs}

    return result

    
def getItemPows1(subj, task, myfreqs, rois='all', tois=[300,1600], numwords=12, removeOutlier=True):
    from scipy.signal import resample
    from eegQC import checkBipDist, removeOutliers, isBadSession
    from localization import get_elec_regions, checkBadElectrodes
    
    df = get_data_index("r1")
    #Find out the sessions, localization, and montage for this subject
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])
    
    perfs_list = []
    startwords = []
    itemPows = []
    itemPows_raw = []
    listPows = []
    listPows_raw = []
    recalls = []
    lists_complete = []
    regions = []
    badelecs = []
    
    for s, sess in enumerate(sessions):
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s])   #reader for loading CML data
        if isBadSession(subj,task,sess):
            continue
        evs = reader.load('task_events')
        evs = evs[evs['eegoffset'] > -1]
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        list_evs = evs[(evs['type']=='TRIAL') & (evs['list']>0)]
        words_first = word_evs[word_evs['serialpos']==1]
        
        # Only use lists of length 12 and also has a recall period
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete_sess = word_evs['list'].unique()[[n==12 for n in nwords]]
        lists_withrec = evs[evs['type']=='REC_START']['list'].unique()
        lists_withend = evs[evs['type']=='REC_END']['list'].unique()
        lists_complete_sess = list(set(lists_withrec).intersection(set(lists_complete_sess)))
        if subj == 'R1342M':
            lists_complete_sess = list(set(lists_withend).intersection(set(lists_complete_sess)))
        if ((subj == 'R1395M') & (task == 'catFR1') & (sess == 1)):
            lists_complete_sess.remove(10)
        word_evs = word_evs[[l in lists_complete_sess for l in word_evs['list']]]
        list_evs = list_evs[[l in lists_complete_sess for l in list_evs['list']]]
        words_first = words_first[[l in lists_complete_sess for l in words_first['list']]]
        words_lists = list(word_evs['list'])
        
        # Get the list performances for complete lists
        #recalls_sess = word_evs[['item_name','recalled','list','serialpos']]
        #perfs_list_sess = [np.sum(recalls_sess[recalls_sess['list']==l]['recalled']) for l in lists_complete_sess]
        #perfs_list.append(perfs_list_sess)
        recalls_sess = pd.DataFrame()
        
        # Get the pairs
        pairs = reader.load('pairs')
        
        # Fix for abnormally distant electrodes
        valids, _ = checkBipDist(reader,pairs,20)
        pairs = pairs[valids]
        
        # Narrow down the pairs further by checking for bad electrodes
        checks = checkBadElectrodes(reader, pairs)
        pairs = pairs[checks]

        # Get localization
        locz = []
        try:
            locz = reader.load('localization')
        except:
            print('No localization file found')
        regions_sess, _,_,_ = get_elec_regions(locz,pairs)
        
        if rois != 'all':
            regionStr = 'ind.corrected.region'
            if 'ind.corrected.region' not in pairs.columns:
                if 'ind.dural.region' in pairs.columns:
                    regionStr = 'ind.dural.region'
                else:
                    regionStr = 'ind.region'
            #regions_sess = np.array(pairs[regionStr])
            pairs = pairs[[r in rois for r in regions_sess]]
            
        # Get the eeg for all words
        badelecs_sess = []
        
        buf = 1200
        eeg = reader.load_eeg(events=word_evs, rel_start=tois[0]-buf, rel_stop=tois[1]+buf, scheme=pairs)
        eeg = eeg.to_mne()
        sr = eeg.info['sfreq']
        
        #Run TF decomposition and log
        power = time_frequency.tfr_morlet(eeg, freqs=myfreqs, n_cycles=5, return_itc=False, average=False)
        powsa = np.log10(power.data)
        del eeg
        del power
        # Remove buffer
        buf_samps = int(round(sr*buf/1000.))
        powsa = powsa[:,:,:,buf_samps:-buf_samps] # event x elec x freq x time

        # To resample to 50Hz
        #resrate = 50
        #resamps = int((tois[1]-tois[0])/1000*sr/resrate)
        #powsa_res = resample(powsa,resamps,axis=3)
        
        # Remove outlier data points (5 stdevs above mean) by converting to nan
        if removeOutlier:
            powsa, badelecs_sess = removeOutliers(powsa,ax=3)
            badelecs.append(badelecs_sess)
                        
        # for each list, choose random subset
        lls = words_first['list'].unique()
        assert(len(powsa)==len(lls)*12)
        startwords_sess = []
        itemPows2 = []
        for ll_ind, ll in enumerate(lls):
            llwords_all = word_evs[word_evs['list']==ll]

            # pick out the sublist
            startword = np.random.randint(1,13-numwords+1)
            endword = startword+numwords-1
            llPows = powsa[(ll_ind*12):12+(ll_ind*12)]
            itemPows2.append(llPows[startword-1:endword])

            # append to data structures
            words_inc = llwords_all[(llwords_all['serialpos'] >= startword) & 
                                    (llwords_all['serialpos'] < startword+numwords)]
            recalls_sess = recalls_sess.append(words_inc[['item_name','recalled','list','serialpos']],ignore_index=True)
            startwords_sess.append(startword)
        
        # average power within each list, over all lists to get list x elec x freq, for nwords number of words
        itemPows2 = np.array(itemPows2)
        listpows_avg_sess = np.nanmean(itemPows2,axis=(1,4)) # average over items within each list and over timepoints

        listpows_z_sess = zscore(listpows_avg_sess)
        itempows_avg_sess = np.nanmean(powsa,axis=3)
        itemPows_raw.append(itempows_avg_sess)
        del powsa
        itempows_z_sess = zscore(itempows_avg_sess)
        #del itempows_avg_sess
        #del listpows_avg_sess
        listPows_raw.append(listpows_avg_sess)
        listPows.append(listpows_z_sess)
        itemPows.append(itempows_z_sess)
        recalls.append(recalls_sess)
        lists_complete.append(lists_complete_sess)
        perfs_list.append([np.sum(recalls_sess[recalls_sess['list']==lnum]['recalled']) for lnum in recalls_sess['list'].unique()])
        startwords.append(startwords_sess)
        regions.append(regions_sess)
        
    result = {'subject':subj, 'task':task, 'pairs':pairs, 'itemPows_raw':itemPows_raw, 'itemPows':itemPows, 'lists':lists_complete, 'listPows_raw':listPows_raw, 'listPows':listPows, 'recalls':recalls, 'perfs_list':perfs_list, 'pairs_regions':regions, 'myfreqs':myfreqs, 'rois':rois, 'numwords':numwords, 'startwords':startwords, 'tois':tois, 'badelecs':badelecs}

    return result


def movingAverage(x, n):
    ret = np.cumsum(x, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n


def getAvgListPows(subj, task, myfreqs, rois='all', buf=1200):
    from eegQC import checkBipDist, removeOutliers, isBadSession
    from localization import get_elec_regions, checkBadElectrodes
    
    df = get_data_index("r1")
    #Find out the sessions, localization, and montage for this subject
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])
    
    perfs_list = []
    avgListPows_raw = []
    recalls = []
    lists_complete = []
    regions = []
    
    # Start session-level analysis
    for s, sess in enumerate(sessions):
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s])
        if isBadSession(subj,task,sess):
            continue
        evs = reader.load('task_events')
        evs = evs[evs['eegoffset'] > -1]
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        
        # Only use lists of length 12 and also has a recall period
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        # make sure each list has 12 items
        lists_complete_sess = word_evs['list'].unique()[[n==12 for n in nwords]]
        # check that serial positions 1 through 12 are all there
        listwords_set = [ word_evs[word_evs['list']==l] for l in lists_complete_sess ]
        allsps = [ np.all(np.array(listwords['serialpos'])==np.arange(1,13)) for listwords in listwords_set ]
        lists_complete_sess = lists_complete_sess[allsps]

        lists_withrec = evs[evs['type']=='REC_START']['list'].unique()
        lists_withend = evs[evs['type']=='REC_END']['list'].unique()
        lists_complete_sess = list(set(lists_withrec).intersection(set(lists_complete_sess)))
        if subj == 'R1342M':
            lists_complete_sess = list(set(lists_withend).intersection(set(lists_complete_sess)))
        if ((subj == 'R1395M') & (task == 'catFR1') & (sess == 1)):
            lists_complete_sess.remove(10)
        word_evs = word_evs[[l in lists_complete_sess for l in word_evs['list']]]
        
        # Get the list performances for complete lists
        recalls_sess = pd.DataFrame()
        
        # Get the pairs
        pairs = reader.load('pairs')
        
        # Fix for abnormally distant electrodes
        mindist = 20 # minimum valid distance in mm for contacts in bipolar
        valids, _ = checkBipDist(reader,pairs,mindist)
        pairs = pairs[valids]
        
        # Narrow down the pairs further by checking for bad electrodes
        checks = checkBadElectrodes(reader, pairs)
        pairs = pairs[checks]

        # Localize the pairs
        locz = []
        try:
            locz = reader.load('localization')
        except:
            print('No localization file found')
        regions_sess, _,_,_ = get_elec_regions(locz,pairs)
        
        if rois != 'all':
            regionStr = 'ind.corrected.region'
            if 'ind.corrected.region' not in pairs.columns:
                if 'ind.dural.region' in pairs.columns:
                    regionStr = 'ind.dural.region'
                else:
                    regionStr = 'ind.region'
            pairs = pairs[[r in rois for r in regions_sess]]
            
        # Get the eeg for all lists
        avgListPows_raw_sess = []
        for ll in lists_complete_sess:
            list_evs = evs[evs['list']==ll]
            list_words = word_evs[word_evs['list']==ll].copy()
            dist_time = int(list_evs[list_evs['type']=='DISTRACT_START']['mstime'])
            lastword_time = int(list_words[list_words['serialpos']==12]['mstime'])
            assert(dist_time-lastword_time >= 1600) # Make sure the distractor task doesn't start too soon after last word

            firstword = list_words[list_words['serialpos']==1]
            firstword_time = int(list_words[list_words['serialpos']==1]['mstime'])
            listtimewindow = lastword_time+1600-firstword_time

            # get EEG
            eeg = reader.load_eeg(events=firstword, rel_start=0-buf, rel_stop=listtimewindow+buf, scheme=pairs)
            eeg = eeg.to_mne()
            tfpower = time_frequency.tfr_morlet(eeg, freqs=myfreqs, n_cycles=5, return_itc=False, average=False)
            powsa = np.log10(tfpower.data).squeeze() # event x elec x freq x time but squeeze out the one event axis
            buf_samps = int(round(eeg.info['sfreq']*buf/1000.))
            pows_avg = np.nanmean(powsa[:,:,buf_samps:-buf_samps],axis=2) # average over time excluding buffer, so now elec x freq
            del eeg, tfpower, powsa

            avgListPows_raw_sess.append(pows_avg)
            del pows_avg
            # add column for real times relative to first word
            list_words['mstime_rel'] = list(list_words['mstime']-firstword_time)
            recalls_sess = recalls_sess.append(list_words[['item_name','recalled','list','serialpos','mstime_rel']],ignore_index=True)

        # Make sure that dimensions match for all lists
        assert(np.all([avgListPows_raw_sess[i].shape==avgListPows_raw_sess[0].shape for i,_ in enumerate(lists_complete_sess)]))
        
        avgListPows_raw.append(np.array(avgListPows_raw_sess))
        recalls.append(recalls_sess)
        lists_complete.append(lists_complete_sess)
        perfs_list.append([np.sum(recalls_sess[recalls_sess['list']==lnum]['recalled']) for lnum in recalls_sess['list'].unique()])
        regions.append(regions_sess)
        
        del avgListPows_raw_sess
        
    result = {'subject':subj, 'task':task, 'pairs':pairs, 'lists':lists_complete, 'avgListPows_raw':avgListPows_raw, 'recalls':recalls, 'perfs_list':perfs_list, 'pairs_regions':regions, 'myfreqs':myfreqs, 'rois':rois}

    return result


def getAvgPreListPows(subj, task, myfreqs, rois='all', toi=[-30000,0], buf=1200):
    from eegQC import checkBipDist, removeOutliers, isBadSession
    from localization import get_elec_regions, checkBadElectrodes
    
    df = get_data_index("r1")
    #Find out the sessions, localization, and montage for this subject
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])
    
    perfs_list = []
    avgListPows_raw = []
    recalls = []
    lists_complete = []
    regions = []
    
    # Start session-level analysis
    for s, sess in enumerate(sessions):
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s])
        if isBadSession(subj,task,sess):
            continue
        evs = reader.load('task_events')
        evs = evs[evs['eegoffset'] > -1]
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        
        # Only use lists of length 12 and also has a recall period
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        # make sure each list has 12 items
        lists_complete_sess = word_evs['list'].unique()[[n==12 for n in nwords]]
        # check that serial positions 1 through 12 are all there
        listwords_set = [ word_evs[word_evs['list']==l] for l in lists_complete_sess ]
        allsps = [ np.all(np.array(listwords['serialpos'])==np.arange(1,13)) for listwords in listwords_set ]
        lists_complete_sess = lists_complete_sess[allsps]

        lists_withrec = evs[evs['type']=='REC_START']['list'].unique()
        lists_withend = evs[evs['type']=='REC_END']['list'].unique()
        lists_complete_sess = list(set(lists_withrec).intersection(set(lists_complete_sess)))
        if subj == 'R1342M':
            lists_complete_sess = list(set(lists_withend).intersection(set(lists_complete_sess)))
        if ((subj == 'R1395M') & (task == 'catFR1') & (sess == 1)):
            lists_complete_sess.remove(10)
        word_evs = word_evs[[l in lists_complete_sess for l in word_evs['list']]]
        firstwords = word_evs[word_evs['serialpos']==1]

        # Get the list performances for complete lists
        recalls_sess = pd.DataFrame()
        
        # Get the pairs
        pairs = reader.load('pairs')
        
        # Fix for abnormally distant electrodes
        mindist = 20 # minimum valid distance in mm for contacts in bipolar
        valids, _ = checkBipDist(reader,pairs,mindist)
        pairs = pairs[valids]
        
        # Narrow down the pairs further by checking for bad electrodes
        checks = checkBadElectrodes(reader, pairs)
        pairs = pairs[checks]

        # Localize the pairs
        locz = []
        try:
            locz = reader.load('localization')
        except:
            print('No localization file found')
        regions_sess, _,_,_ = get_elec_regions(locz,pairs)
        
        if rois != 'all':
            regionStr = 'ind.corrected.region'
            if 'ind.corrected.region' not in pairs.columns:
                if 'ind.dural.region' in pairs.columns:
                    regionStr = 'ind.dural.region'
                else:
                    regionStr = 'ind.region'
            pairs = pairs[[r in rois for r in regions_sess]]
            
        # Get the eeg for all lists and average over time of interest
        eeg = reader.load_eeg(events=firstwords, rel_start=toi[0]-buf-buf, rel_stop=toi[1]-buf+buf, scheme=pairs)
        eeg = eeg.to_mne()
        tfpower = time_frequency.tfr_morlet(eeg, freqs=myfreqs, n_cycles=5, return_itc=False, average=False)
        powsa = np.log10(tfpower.data) # event x elec x freq x time but squeeze out the one event axis
        buf_samps = int(round(eeg.info['sfreq']*buf/1000.))
        pows_avg = np.nanmean(powsa[:,:,:,buf_samps:-buf_samps],axis=3) # average over time excluding buffer, so now elec x freq
        del eeg, tfpower, powsa

        # append the behavioral data of interest
        for ll in lists_complete_sess:
            list_evs = evs[evs['list']==ll]
            list_words = word_evs[word_evs['list']==ll].copy()

            firstword = list_words[list_words['serialpos']==1]
            firstword_time = int(list_words[list_words['serialpos']==1]['mstime'])
            
            # add column for real times relative to first word
            list_words['mstime_rel'] = list(list_words['mstime']-firstword_time)
            recalls_sess = recalls_sess.append(list_words[['item_name','recalled','list','serialpos','mstime_rel']],ignore_index=True)

        avgListPows_raw.append(pows_avg)
        del pows_avg
        recalls.append(recalls_sess)
        lists_complete.append(lists_complete_sess)
        perfs_list.append([np.sum(recalls_sess[recalls_sess['list']==lnum]['recalled']) for lnum in recalls_sess['list'].unique()])
        regions.append(regions_sess)
        
        
    result = {'subject':subj, 'task':task, 'pairs':pairs, 'lists':lists_complete, 'avgListPows_raw':avgListPows_raw, 'recalls':recalls, 'perfs_list':perfs_list, 'pairs_regions':regions, 'myfreqs':myfreqs, 'rois':rois}

    return result


def getWholeListPows1(subj, task, myfreqs, rois='all', tois=[-1000,31000], numwords=12, removeOutlier=False, winsize=1000, increment=100):
    from scipy.signal import resample
    from eegQC import checkBipDist, removeOutliers, isBadSession
    from localization import get_elec_regions, checkBadElectrodes
    from scipy.interpolate import interp1d
    
    df = get_data_index("r1")
    #Find out the sessions, localization, and montage for this subject
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])
    
    perfs_list = []
    startwords = []
    wholeListPows_raw = []
    recalls = []
    lists_complete = []
    regions = []
    badelecs = []
    
    # Fix timing stuff
    #winsize = 1000
    #increment = 100
    buf = 1200
    sr_new = 500
    sr_ms = sr_new/1000.
    buf_samps = int(round(sr_new*buf/1000.))
    times_standard = np.linspace(tois[0],tois[1],1+int(sr_new/1000*np.diff(tois)))[:-1] # standard being with sample rate of 500Hz
    times_standard = np.array([round(t,1) for t in times_standard]) # round to 0.1ms
    
    winsize_samps = int(sr_ms*winsize)
    increment_samps = int(sr_ms*increment)
    # Take every 100ms (or whatever increment is)
    times_final = times_standard[winsize_samps-1:][::increment_samps] - int(winsize/2 - 1/sr_ms)
    
    # Start session-level analysis
    for s, sess in enumerate(sessions):
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s]) #reader for loading CML data
        if isBadSession(subj,task,sess):
            continue
        evs = reader.load('task_events')
        evs = evs[evs['eegoffset'] > -1]
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        list_evs = evs[(evs['type']=='TRIAL') & (evs['list']>0)]
        words_first = word_evs[word_evs['serialpos']==1]
        
        # Only use lists of length 12 and also has a recall period
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete_sess = word_evs['list'].unique()[[n==12 for n in nwords]]
        lists_withrec = evs[evs['type']=='REC_START']['list'].unique()
        lists_withend = evs[evs['type']=='REC_END']['list'].unique()
        lists_complete_sess = list(set(lists_withrec).intersection(set(lists_complete_sess)))
        if subj == 'R1342M':
            lists_complete_sess = list(set(lists_withend).intersection(set(lists_complete_sess)))
        if ((subj == 'R1395M') & (task == 'catFR1') & (sess == 1)):
            lists_complete_sess.remove(10)
        word_evs = word_evs[[l in lists_complete_sess for l in word_evs['list']]]
        list_evs = list_evs[[l in lists_complete_sess for l in list_evs['list']]]
        words_first = words_first[[l in lists_complete_sess for l in words_first['list']]]
        words_lists = list(word_evs['list'])
        
        # Get the list performances for complete lists
        #recalls_sess = word_evs[['item_name','recalled','list','serialpos']]
        #perfs_list_sess = [np.sum(recalls_sess[recalls_sess['list']==l]['recalled']) for l in lists_complete_sess]
        #perfs_list.append(perfs_list_sess)
        recalls_sess = pd.DataFrame()
        
        # Get the pairs
        pairs = reader.load('pairs')
        
        # Fix for abnormally distant electrodes
        valids, _ = checkBipDist(reader,pairs,20)
        pairs = pairs[valids]
        
        # Narrow down the pairs further by checking for bad electrodes
        checks = checkBadElectrodes(reader, pairs)
        pairs = pairs[checks]

        # Get localization
        locz = []
        try:
            locz = reader.load('localization')
        except:
            print('No localization file found')
        regions_sess, _,_,_ = get_elec_regions(locz,pairs)
        
        if rois != 'all':
            regionStr = 'ind.corrected.region'
            if 'ind.corrected.region' not in pairs.columns:
                if 'ind.dural.region' in pairs.columns:
                    regionStr = 'ind.dural.region'
                else:
                    regionStr = 'ind.region'
            #regions_sess = np.array(pairs[regionStr])
            pairs = pairs[[r in rois for r in regions_sess]]
            
        # Get the eeg for all words
        badelecs_sess = []
        
        eeg = reader.load_eeg(events=words_first, rel_start=tois[0]-buf, rel_stop=tois[1]+buf, scheme=pairs)
        eeg = eeg.to_mne()
        times_sess = eeg.times*1000 # to get in ms
        #sr = eeg.info['sfreq']
        
        #Run TF decomposition and log
        power = time_frequency.tfr_morlet(eeg, freqs=myfreqs, n_cycles=5, return_itc=False, average=False)
        powsa = np.log10(power.data)
        del eeg
        del power
        
        assert(powsa.shape[-1] == len(times_sess))
        
        powsb = np.zeros((powsa.shape[0],powsa.shape[1],powsa.shape[2],len(times_standard)))
        # Resample by interpolation, to sr_new, and remove buffer at the same time
        for f, freq in enumerate(myfreqs):
            fpow = interp1d(times_sess,powsa[:,:,f,:],kind='linear',axis=-1) # define the function from time to power
            powsb[:,:,f,:] = fpow(times_standard) # interpolate to the standard time axis, minus the buffer zone
        powsa = powsb.copy()
        del powsb
        
        # Resample by scipy.signal function
        #powsa = 
        
        # Remove buffer
        #powsa = powsa[:,:,:,buf_samps:-buf_samps] # event x elec x freq x time
        #times_sess = times_sess[buf_samps:-buf_samps]
        
        # Remove outlier data points (5 stdevs above mean) by converting to nan
        if removeOutlier:
            powsa, badelecs_sess = removeOutliers(powsa,ax=3)
            badelecs.append(badelecs_sess)
                               
        # for each list, create sliding window averages of 1000ms, sliding by 100ms each
        cumsum = np.cumsum(powsa, axis=3, dtype=float)
        cumsum[:,:,:,winsize_samps:] = cumsum[:,:,:,winsize_samps:] - cumsum[:,:,:,:-winsize_samps]
        wholeListPows_raw_sess = cumsum[:,:,:,winsize_samps - 1:] / winsize_samps 
        wholeListPows_raw_sess = wholeListPows_raw_sess[:,:,:,::increment_samps] # just take the ones every 100ms
        #times_sess = times_sess[winsize_samps-1:][::increment_samps] - int(winsize/2 - 1/sr_ms)
        #times.append(times_sess)
        del powsa, cumsum
        
        lls = words_first['list'].unique()
        assert(len(wholeListPows_raw_sess)==len(lls))
        startwords_sess = []
        for ll_ind, ll in enumerate(lls):
            llwords_all = word_evs[word_evs['list']==ll]

            # pick out the sublist
            startword = np.random.randint(1,13-numwords+1)
            endword = startword+numwords-1

            # append to data structures
            words_inc = llwords_all[(llwords_all['serialpos'] >= startword) & 
                                    (llwords_all['serialpos'] < startword+numwords)]
            # add column for real times relative to first word
            words_inc['mstime_rel'] = list(llwords_all['mstime']-llwords_all.iloc()[0]['mstime'])
            recalls_sess = recalls_sess.append(words_inc[['item_name','recalled','list','serialpos','mstime_rel']],ignore_index=True)
            startwords_sess.append(startword)
        
        # z score the time series - subtract the list means from each window, and divide by variance 
        #sessmean = np.nanmean(wholeListPows_raw,axis=(0,3)) # elec x freq
        #sessmean = np.expand_dims(sessmean,(0,3))
        #sessmean = np.repeat(np.repeat(sessmean,wholeListPows_raw.shape[0],axis=0),wholeListPows_raw.shape[3],axis=3)
        
        #sessvar = np.nanstd(wholeListPows_raw,axis=(0,3)) # elec x freq
        #sessvar = np.expand_dims(sessvar,(0,3))
        #sessvar = np.repeat(np.repeat(sessvar,wholeListPows_raw.shape[0],axis=0),wholeListPows_raw.shape[3],axis=3)

        #wholeListPows = np.subtract(wholeListPows_raw,sessmean)/sessvar
        wholeListPows_raw.append(wholeListPows_raw_sess)
        recalls.append(recalls_sess)
        lists_complete.append(lists_complete_sess)
        perfs_list.append([np.sum(recalls_sess[recalls_sess['list']==lnum]['recalled']) for lnum in recalls_sess['list'].unique()])
        startwords.append(startwords_sess)
        regions.append(regions_sess)
        
        del wholeListPows_raw_sess
        
    result = {'subject':subj, 'task':task, 'pairs':pairs, 'lists':lists_complete, 'wholeListPows_raw':wholeListPows_raw, 'recalls':recalls, 'perfs_list':perfs_list, 'pairs_regions':regions, 'myfreqs':myfreqs, 'rois':rois, 'numwords':numwords, 'startwords':startwords, 'tois':tois, 'times_standard':times_standard, 'times_final':times_final, 'badelecs':badelecs}

    return result


def getItemPows_slidingwindow(subj, task, myfreqs, rois='all', tois=[-1000,1000], removeOutlier=False):
    from scipy.signal import resample
    from eegQC import checkBipDist, removeOutliers, isBadSession
    from localization import get_elec_regions, checkBadElectrodes
    from scipy.interpolate import interp1d
    
    df = get_data_index("r1")
    #Find out the sessions, localization, and montage for this subject
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])
    
    perfs_list = []
    itemPows_raw = []
    recalls = []
    lists_complete = []
    regions = []
    badelecs = []
    
    # Fix timing stuff
    winsize = 1000
    increment = 100
    buf = 1200
    sr_new = 500
    sr_ms = sr_new/1000.
    buf_samps = int(round(sr_new*buf/1000.))
    times_standard = np.linspace(tois[0],tois[1],1+int(sr_new/1000*np.diff(tois)))[:-1] # standard being with sample rate of 500Hz
    times_standard = np.array([round(t,1) for t in times_standard]) # round to 0.1ms
    
    winsize_samps = int(sr_ms*winsize)
    increment_samps = int(sr_ms*increment)
    # Take every 100ms (or whatever increment is)
    times_final = times_standard[winsize_samps-1:][::increment_samps] - int(winsize/2 - 1/sr_ms)
    
    # Start session-level analysis
    for s, sess in enumerate(sessions):
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s])   #reader for loading CML data
        if isBadSession(subj,task,sess):
            continue
        evs = reader.load('task_events')
        evs = evs[evs['eegoffset'] > -1]
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        list_evs = evs[(evs['type']=='TRIAL') & (evs['list']>0)]
        words_first = word_evs[word_evs['serialpos']==1]
        
        # Only use lists of length 12 and also has a recall period
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete_sess = word_evs['list'].unique()[[n==12 for n in nwords]]
        lists_withrec = evs[evs['type']=='REC_START']['list'].unique()
        lists_withend = evs[evs['type']=='REC_END']['list'].unique()
        lists_complete_sess = list(set(lists_withrec).intersection(set(lists_complete_sess)))
        if subj == 'R1342M':
            lists_complete_sess = list(set(lists_withend).intersection(set(lists_complete_sess)))
        if ((subj == 'R1395M') & (task == 'catFR1') & (sess == 1)):
            lists_complete_sess.remove(10)
        lists_complete.append(lists_complete_sess)
        word_evs = word_evs[[l in lists_complete_sess for l in word_evs['list']]]
        list_evs = list_evs[[l in lists_complete_sess for l in list_evs['list']]]
        words_first = words_first[[l in lists_complete_sess for l in words_first['list']]]
        words_lists = list(word_evs['list'])
        
        # Get the list performances for complete lists
        recalls_sess = word_evs[['item_name','recalled','list','serialpos']]
        recalls.append(recalls_sess)
        #perfs_list_sess = [np.sum(recalls_sess[recalls_sess['list']==l]['recalled']) for l in lists_complete_sess]
        #perfs_list.append(perfs_list_sess)
        #recalls_sess = pd.DataFrame()
        
        # Get the pairs
        pairs = reader.load('pairs')
        
        # Fix for abnormally distant electrodes
        valids, _ = checkBipDist(reader,pairs,20)
        pairs = pairs[valids]
        
        # Narrow down the pairs further by checking for bad electrodes
        checks = checkBadElectrodes(reader, pairs)
        pairs = pairs[checks]

        # Get localization
        locz = []
        try:
            locz = reader.load('localization')
        except:
            print('No localization file found')
        regions_sess, _,_,_ = get_elec_regions(locz,pairs)
        regions.append(regions_sess)
        
        if rois != 'all':
            regionStr = 'ind.corrected.region'
            if 'ind.corrected.region' not in pairs.columns:
                if 'ind.dural.region' in pairs.columns:
                    regionStr = 'ind.dural.region'
                else:
                    regionStr = 'ind.region'
            pairs = pairs[[r in rois for r in regions_sess]]
            
        # Get the eeg for all words
        badelecs_sess = []
        
        ###################
        # Now get the individual item power
        eeg = reader.load_eeg(events=word_evs, rel_start=tois[0]-buf, rel_stop=tois[1]+buf, scheme=pairs)
        eeg = eeg.to_mne()
        times_sess = eeg.times*1000 # to get in ms
        #sr = eeg.info['sfreq']
        sr_ms = sr_new/1000.
        
        #Run TF decomposition and log
        power = time_frequency.tfr_morlet(eeg, freqs=myfreqs, n_cycles=5, return_itc=False, average=False)
        powsa = np.log10(power.data)
        del eeg, power
        
        # Resample by interpolation, to sr_new, and remove buffer at the same time
        assert(powsa.shape[-1] == len(times_sess))
        powsb = np.zeros((powsa.shape[0],powsa.shape[1],powsa.shape[2],len(times_standard)))
        for f, freq in enumerate(myfreqs):
            fpow = interp1d(times_sess,powsa[:,:,f,:],kind='linear',axis=-1) # define the function from time to power
            powsb[:,:,f,:] = fpow(times_standard) # interpolate to the standard time axis, minus the buffer zone
        powsa = powsb.copy()
        del powsb
        # Remove buffer
        #powsa = powsa[:,:,:,buf_samps:-buf_samps] # event x elec x freq x time
        #times_sess = times_sess[buf_samps:-buf_samps]
        
        # Remove outliers
        if removeOutlier:
            powsa, badelecs_sess = removeOutliers(powsa,ax=3)
            badelecs.append(badelecs_sess)

        assert(len(powsa)==len(words_first['list'].unique())*12)
        
        # Do sliding windows
        winsize_samps = int(sr_ms*winsize)
        increment_samps = int(sr_ms*increment)
        #numwins = int(len(times)/increment_samps)
        #timewins_samps = [increment_samps*i for i in range(numwins)]
        
        # for each list, create sliding window averages of 1000ms, sliding by 100ms each
        cumsum = np.cumsum(powsa, axis=3, dtype=float)
        cumsum[:,:,:,winsize_samps:] = cumsum[:,:,:,winsize_samps:] - cumsum[:,:,:,:-winsize_samps]
        itemPows_raw_sess = cumsum[:,:,:,winsize_samps - 1:] / winsize_samps 
        itemPows_raw_sess = itemPows_raw_sess[:,:,:,::increment_samps] # just take the ones every 50 samples (i.e. 100ms)
        #times_sess = times_sess[winsize_samps-1:][::increment_samps] - int(winsize/2 - 1/sr_ms)
        #times.append(times_sess)
        del cumsum, powsa
        
        # alternatively, force into specific number of windows
        #numwins = int((np.diff(tois)[0]-winsize)/increment + 1)
        #times = int(winsize/2) + np.linspace(tois[0],tois[1]-winsize,numwins)
        #winsize_samps = 
        #increment_samps = int(powsa.shape[-1]/numwins)
        
        itemPows_raw.append(itemPows_raw_sess)
        del itemPows_raw_sess
        
        perfs_list.append([np.sum(recalls_sess[recalls_sess['list']==lnum]['recalled']) for lnum in recalls_sess['list'].unique()])

        
    result = {'subject':subj, 'task':task, 'pairs':pairs, 'lists':lists_complete, 'itemPows_raw':itemPows_raw, 'recalls':recalls, 'perfs_list':perfs_list, 'pairs_regions':regions, 'myfreqs':myfreqs, 'rois':rois, 'tois':tois, 'times_standard':times_standard, 'times_final':times_final, 'badelecs':badelecs}

    return result


def getListPows(subj, task, myfreqs, rois='all', tois=[300,1600]):
    from scipy.signal import resample
    from eegQC import checkBipDist
    
    df = get_data_index("r1")
    #Find out the sessions, localization, and montage for this subject
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])
    
    perfs_list = []
    listPows = []
    n_outliers = []
    
    for s, sess in enumerate(sessions):
        # For first session...
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s])   #reader for loading CML data
        evs = reader.load('task_events')
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        list_evs = evs[(evs['type']=='TRIAL') & (evs['list']>0)]
        words_first = word_evs[word_evs['serialpos']==1]
        
        # Only use lists of length 12
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete = word_evs['list'].unique()[[n==12 for n in nwords]]
        word_evs = word_evs[[l in lists_complete for l in word_evs['list']]]
        list_evs = list_evs[[l in lists_complete for l in list_evs['list']]]
        words_first = words_first[[l in lists_complete for l in words_first['list']]]
        words_lists = list(word_evs['list'])
        
        # Get the list performances for complete lists
        recalls = word_evs[['recalled','list']]
        perfs_list_sess = [np.sum(recalls[recalls['list']==l]['recalled']) for l in lists_complete]
        perfs_list.append(perfs_list_sess)
        
        # Get the pairs
        pairs = reader.load('pairs')
        
        # Fix for abnormally distant electrodes
        valids, _ = checkBipDist(reader,pairs,20)
        pairs = pairs[valids]
        
        # Narrow down the pairs further by checking for bad electrodes
        checks = checkBadElectrodes(reader, pairs)
        pairs = pairs[checks]
        
        # Narrow down the pairs
        if rois != 'all':
            regionStr = 'ind.corrected.region'
            if 'ind.corrected.region' not in pairs.columns:
                if 'ind.dural.region' in pairs.columns:
                    regionStr = 'ind.dural.region'
                else:
                    regionStr = 'ind.region'
            regions = np.array(pairs[regionStr])
            pairs = pairs[[r in rois for r in regions]]
            
        # Get the eeg for all lists
        n_outliers_sess = []
        buf = 1200
        eeg = reader.load_eeg(events=word_evs, rel_start=tois[0]-buf, rel_stop=tois[1]+buf, scheme=pairs)
        eeg = eeg.to_mne()
        sr = eeg.info['sfreq']
        
        #Run TF decomposition and log
        power = time_frequency.tfr_morlet(eeg, freqs=myfreqs, n_cycles=6, return_itc=False, average=False)
        powsa = np.log10(power.data)
        del eeg
        del power
        # Remove buffer
        buf_samps = int(round(sr*buf/1000.))
        powsa = powsa[:,:,:,buf_samps:-buf_samps] # event x elec x freq x time

        # To resample to 50Hz
        #resrate = 50
        #resamps = int((tois[1]-tois[0])/1000*sr/resrate)
        #powsa_res = resample(powsa,resamps,axis=3)
        
        # Remove outlier data points (6 stdevs above mean) by converting to nan
        maxval = np.mean(powsa,3)+6*np.std(powsa,3)
        maxval = np.expand_dims(maxval,3)
        maxval = np.repeat(maxval,powsa.shape[3],axis=3)
        n_outliers_sess.append(np.sum(powsa > maxval))
        powsa[powsa > maxval]=np.nan
        
        # Average power across frequencies
        #powsa_avg = np.nanmean(powsa,axis=2)
        
        # Average power within each list, over all lists to get list x elec x freq
        listpows_avg_sess = np.array([np.nanmean(powsa[words_lists == l],axis=(0,3)) for l in np.unique(words_lists)])
        del powsa
        # Then z-score the overall list powers
        listpows_z_sess = zscore(listpows_avg_sess)
        del listpows_avg_sess
        listPows.append(listpows_z_sess)
        n_outliers.append(n_outliers_sess)
    
    result = {'pairs':pairs, 'listPows':listPows, 'perfs_list':perfs_list, 'myfreqs':myfreqs, 'rois':rois, 'tois':tois, 'n_outliers':n_outliers}

    return result


def getWholeListPows(subj, task, myfreqs, rois='all', numwords=12):
    from scipy.signal import resample
    from eegQC import checkBipDist, removeOutliers
    from localization import get_elec_regions
    
    df = get_data_index("r1")
    #Find out the sessions, localization, and montage for this subject
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])
    
    perfs_list = []
    listPows = []
    offsets = []
    badelecs = []
    adjs = []
    regions = []
    recalls = []
    timelengths = []
    
    for s, sess in enumerate(sessions):
        # For first session...
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s])   #reader for loading CML data
        evs = reader.load('task_events')
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        list_evs = evs[(evs['type']=='TRIAL') & (evs['list']>0)]
        
        
        # Only use lists of length 12
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete = word_evs['list'].unique()[[n==12 for n in nwords]]
        word_evs = word_evs[[l in lists_complete for l in word_evs['list']]]
        list_evs = list_evs[[l in lists_complete for l in list_evs['list']]]
        nlists = len(word_evs['list'].unique())
        
        # Generate the sublists - 12/numwords sublists of length numwords
        startwords = np.arange(12)[::numwords]
        numsegs = len(startwords)
        sublists = np.zeros(len(word_evs),dtype=int)
        for n in range(numsegs):
            a=[(word_evs.iloc()[w]['serialpos']>=startwords[n]+1) & (word_evs.iloc()[w]['serialpos']<startwords[n]+1+numwords) for w in range(len(word_evs))]
            sublists = sublists+(n+1)*np.array(a,dtype=int)
        
        sublists = numsegs*(np.array(word_evs['list'])-1)+sublists
        subsps = np.tile(np.arange(1,numwords+1),nlists*numsegs)
        
        word_evs['sublist'] = sublists
        word_evs['subserialpos'] = subsps
        words_first = word_evs[word_evs['subserialpos']==1]
        
        # Get the list performances for sublists
        sublists_complete = word_evs['sublist'].unique()
        recalls_sess = word_evs[['item_name','recalled','list','serialpos','sublist','subserialpos','mstime']]
        perfs_list_sess = [np.sum(recalls_sess[recalls_sess['sublist']==sl]['recalled']) for sl in sublists_complete]
        perfs_list.append(perfs_list_sess)

        # Get the pairs
        pairs = reader.load('pairs')
        
        # Fix for abnormally distant electrodes
        valids, _ = checkBipDist(reader,pairs,20) #20mm threshold
        pairs = pairs[valids]
        
        # Narrow down the pairs further by checking for bad electrodes
        checks = checkBadElectrodes(reader, pairs)
        pairs = pairs[checks]
        
        # Get localization using John's method
        locz = []
        try:
            locz = reader.load('localization')
        except:
            print('No localization file found')
        regions_sess, _,_,_ = get_elec_regions(locz,pairs)

        # Narrow down the pairs
        if rois != 'all':
            regionStr = 'ind.corrected.region'
            if 'ind.corrected.region' not in pairs.columns:
                if 'ind.dural.region' in pairs.columns:
                    regionStr = 'ind.dural.region'
                else:
                    regionStr = 'ind.region'
            pairs = pairs[[r in rois for r in regions_sess]]
            
        # Get the eeg for all lists
        myfreqs = [int(round(x)) for x in np.logspace(np.log10(3),np.log10(200),12)]
        pows_sess = []
        badelecs_sess = []
        offsets_sess = []
        timelen_sess = []
        adjs_sess = []
        
        lls = words_first['sublist'].unique()

        start = 300

        # For each sublist
        for ll in lls:
            # Get list words timings
            llwords_first = words_first[words_first['sublist']==ll]
            llwords_all = word_evs[word_evs['sublist']==ll]
            llwords_reltimes = np.array([t-int(llwords_first['mstime']) for t in list(llwords_all['mstime'])])
            off_time = llwords_reltimes[-1]+1600
            timelen_sess.append(off_time)
                        
            # Get the EEG
            buf = 1200
            eeg = reader.load_eeg(events=llwords_first, rel_start=start-buf, rel_stop=start+off_time+buf, scheme=pairs)
            eeg_mne = eeg.to_mne()
            #Fix EEG length
            eeg, eeg_mne, adj = fixMNEoffset(eeg,eeg_mne,reader,llwords_first,start-buf,start+off_time+buf,pairs)
            adjs_sess.append(adj)
            if np.isnan(adj):
                continue
            sr = eeg_mne.info['sfreq']

            #Run TF decomposition and log
            power = time_frequency.tfr_morlet(eeg_mne, freqs=myfreqs, n_cycles=6, return_itc=False, average=False)
            powsa = np.log10(power.data)
            del eeg
            del eeg_mne
            del power
            # Remove buffer
            buf_samps = int(round(sr*buf/1000.))
            powsa = powsa[:,:,:,buf_samps:-buf_samps] # event x elec x freq x time

            # To resample to 50Hz
            #resrate = 50
            #resamps = int((tois[1]-tois[0])/1000*sr/resrate)
            #powsa_res = resample(powsa,resamps,axis=3)
            
            # Remove outlier data points (6 stdevs above mean) by converting to nan
            powsa, badelecs_ll = removeOutliers(powsa,ax=3)
            badelecs_sess.append(badelecs_ll)
            # Average across time
            powsa_avg = np.nanmean(powsa,axis=3) # now event x elec x freq
            pows_sess.append(powsa_avg)
            del powsa
            del powsa_avg
        
        # Z-score power
        listpows_z_sess = zscore(pows_sess,axis=0) 
        listpows_z_sess = np.squeeze(listpows_z_sess,axis=1) # this will now be list x electrode x frequency
        listPows.append(listpows_z_sess)
        recalls.append(recalls_sess)
        timelengths.append(timelen_sess)
        regions.append(regions_sess)
        adjs.append(adjs_sess)
        badelecs.append(badelecs_sess)
        
    result = {'pairs':pairs, 'pairs_regions':regions, 'listPows':listPows, 'recalls':recalls, 'perfs_list':perfs_list, 'myfreqs':myfreqs, 'rois':rois, 'numwords':numwords, 'timelengths':timelengths, 'badelecs':badelecs, 'adjs':adjs}

    return result


def getWholeListPows0(subj, task, myfreqs, rois='all', numwords=12):
    from scipy.signal import resample
    from eegQC import checkBipDist, removeOutliers
    from localization import get_elec_regions
    
    df = get_data_index("r1")
    #Find out the sessions, localization, and montage for this subject
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])
    
    perfs_list = []
    listPows = []
    offsets = []
    startwords = []
    badelecs = []
    adjs = []
    regions = []
    recalls = []
    timelengths = []
    
    for s, sess in enumerate(sessions):
        # For first session...
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s])   #reader for loading CML data
        evs = reader.load('task_events')
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        list_evs = evs[(evs['type']=='TRIAL') & (evs['list']>0)]
        words_first = word_evs[word_evs['serialpos']==1]
        
        # Only use lists of length 12
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete = word_evs['list'].unique()[[n==12 for n in nwords]]
        word_evs = word_evs[[l in lists_complete for l in word_evs['list']]]
        list_evs = list_evs[[l in lists_complete for l in list_evs['list']]]
        words_first = words_first[[l in lists_complete for l in words_first['list']]]
        words_lists = list(word_evs['list'])
        
        # Get the list performances for complete lists
        #recalls_sess = word_evs[['item_name','recalled','list','serialpos']]
        #perfs_list_sess = [np.sum(recalls_sess[recalls_sess['list']==l]['recalled']) for l in lists_complete]
        #perfs_list.append(perfs_list_sess)
        recalls_sess = pd.DataFrame()

        # Get the pairs
        pairs = reader.load('pairs')
        
        # Fix for abnormally distant electrodes
        valids, _ = checkBipDist(reader,pairs,20)
        pairs = pairs[valids]
        
        # Narrow down the pairs further by checking for bad electrodes
        checks = checkBadElectrodes(reader, pairs)
        pairs = pairs[checks]

        # Get localization using John's method
        locz = []
        try:
            locz = reader.load('localization')
        except:
            print('No localization file found')
        regions_sess, _,_,_ = get_elec_regions(locz,pairs)

        # Narrow down the pairs
        if rois != 'all':
            regionStr = 'ind.corrected.region'
            if 'ind.corrected.region' not in pairs.columns:
                if 'ind.dural.region' in pairs.columns:
                    regionStr = 'ind.dural.region'
                else:
                    regionStr = 'ind.region'
            #regions_sess = np.array(pairs[regionStr])
            pairs = pairs[[r in rois for r in regions_sess]]
            
        # Get the eeg for all lists
        myfreqs = [int(round(x)) for x in np.logspace(np.log10(3),np.log10(200),12)]
        pows_sess = []
        badelecs_sess = []
        offsets_sess = []
        timelen_sess = []
        startwords_sess = []
        adjs_sess = []
        lls = words_first['list'].unique()

        start = 0
        end = 30000
        # Randomly choose interval of interest, of length toi, within whole list, different offset for each list, BY WORD, not time
        for ll in lls:
            llwords = words_first[words_first['list']==ll]
            llwords_all = word_evs[word_evs['list']==ll]
            llwords_reltimes = np.array([t-int(llwords['mstime']) for t in list(llwords_all['mstime'])])
            #offset_per = np.random.random()
            #range_start = [start,end-toi] #range of permissible start time values for the window of interest
            #offset_time = int(offset_per*np.diff(range_start)[0])
            #offsets_sess.append(offset_time)

            # Get performance for this interval of the list
            # Decide which words to include based on following criteria
            #includetimes = [(t+1600>offset_time) & (t+950<offset_time+toi) for t in llwords_reltimes]
            #words_inc = llwords_all[includetimes]
            #recalls_sess = recalls_sess.append(words_inc[['item_name','recalled','list','serialpos']],ignore_index=True)

            #to start at word instead of time (300ms post word1 presentation)
            startword = np.random.randint(1,13-numwords+1)
            on_time = llwords_reltimes[startword-1]+300
            off_time = llwords_reltimes[startword+numwords-2]+1600
            words_inc = llwords_all[(llwords_all['serialpos'] >= startword) & (llwords_all['serialpos'] < startword+numwords)]
            recalls_sess = recalls_sess.append(words_inc[['item_name','recalled','list','serialpos']],ignore_index=True)
            timelen_sess.append(off_time-on_time)
            startwords_sess.append(startword)
            
            # Get the EEG
            buf = 1200
            #eeg = reader.load_eeg(events=llwords, rel_start=start+offset_time-buf, rel_stop=start+offset_time+toi+buf, scheme=pairs)
            eeg = reader.load_eeg(events=llwords, rel_start=start+on_time-buf, rel_stop=start+off_time+buf, scheme=pairs)
            eeg_mne = eeg.to_mne()
            #Fix EEG length
            eeg, eeg_mne, adj = fixMNEoffset(eeg,eeg_mne,reader,llwords,start+on_time-buf,start+off_time+buf,pairs)
            adjs_sess.append(adj)
            if np.isnan(adj):
                continue
            sr = eeg_mne.info['sfreq']

            #Run TF decomposition and log
            power = time_frequency.tfr_morlet(eeg_mne, freqs=myfreqs, n_cycles=6, return_itc=False, average=False)
            powsa = np.log10(power.data)
            del eeg
            del eeg_mne
            del power
            # Remove buffer
            buf_samps = int(round(sr*buf/1000.))
            powsa = powsa[:,:,:,buf_samps:-buf_samps] # event x elec x freq x time

            # To resample to 50Hz
            #resrate = 50
            #resamps = int((tois[1]-tois[0])/1000*sr/resrate)
            #powsa_res = resample(powsa,resamps,axis=3)
            
            # Remove outlier data points (6 stdevs above mean) by converting to nan
            powsa, badelecs_ll = removeOutliers(powsa,ax=3)
            badelecs_sess.append(badelecs_ll)
            # Average across time
            powsa_avg = np.nanmean(powsa,axis=3)
            pows_sess.append(powsa_avg)
            del powsa
            del powsa_avg
        
        # Z-score power
        listpows_z_sess = zscore(pows_sess,axis=0) 
        listpows_z_sess = np.squeeze(listpows_z_sess,axis=1) # this will now be list x electrode x frequency
        listPows.append(listpows_z_sess)
        offsets.append(offsets_sess)
        startwords.append(startwords_sess)
        recalls.append(recalls_sess)
        timelengths.append(timelen_sess)
        perfs_list.append([np.sum(recalls_sess[recalls_sess['list']==lnum]['recalled']) for lnum in recalls_sess['list'].unique()])
        regions.append(regions_sess)
        adjs.append(adjs_sess)
        badelecs.append(badelecs_sess)
        
    result = {'pairs':pairs, 'pairs_regions':regions, 'listPows':listPows, 'recalls':recalls, 'perfs_list':perfs_list, 'myfreqs':myfreqs, 'rois':rois, 'numwords':numwords, 'timelengths':timelengths, 'startwords':startwords, 'badelecs':badelecs, 'adjs':adjs}

    return result


def getListPowCorr(subj, task, myfreqs, rois='all', tois=[0,30000]):
    #The "r1" database corresponds to all of the RAM subjects
    df = get_data_index("r1")

    #sub = 'R1111M'
    #task = 'FR1'
    #tois = [0,30000]
    #tint = 3000
    times = np.diff(tois)[0]/2

    #Find out the sessions, localization, and montage for this subject
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])

    rs_z = []
    ps = []
    ps_fdr = []
    perfs_list = []
    perfs_z = []
    pows_z = []
    
    for s, sess in enumerate(sessions):
        # For first session...
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s])   #reader for loading CML data
        evs = reader.load('task_events')
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        recalls = word_evs[['recalled','list']]
        list_evs = evs[(evs['type']=='TRIAL') & (evs['list']>0)]
        words_first = word_evs[word_evs['serialpos']==1]
        
        # Only use lists of length 12
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete = word_evs['list'].unique()[[n==12 for n in nwords]]
        word_evs = word_evs[[l in lists_complete for l in word_evs['list']]]
        list_evs = list_evs[[l in lists_complete for l in list_evs['list']]]
        words_first = words_first[[l in lists_complete for l in words_first['list']]]

        # Get the pairs
        pairs = reader.load('pairs')

        # Narrow down the pairs
        if rois != 'all':
            regionStr = 'ind.corrected.region'
            if 'ind.corrected.region' not in pairs.columns:
                if 'ind.dural.region' in pairs.columns:
                    regionStr = 'ind.dural.region'
                else:
                    regionStr = 'ind.region'
            regions = np.array(pairs[regionStr])
            pairs = pairs[[r in rois for r in regions]]
            
        # Get the list performances for complete lists
        perfs_list_sess = [np.sum(recalls[recalls['list']==l]['recalled']) for l in lists_complete]
        perfs_list.append(perfs_list_sess)
        perfs_z.append(zscore(perfs_list_sess))
        # Get the contacts in frontal cortex
        # pairs_frontal, pairs_temporal, pairs_parietal = getRegions(reader, pairs)

        # Get the eeg for all lists
        buf = 1000
        eeg = reader.load_eeg(events=words_first, rel_start=tois[0]-buf, rel_stop=tois[1]+buf, scheme=pairs)
        eeg = eeg.to_mne()
        sr = eeg.info['sfreq']
        
        # Get the band-passed power
        #Pre-define an array of frequencies. We can log-space them since wavelets are less sensitive to differences at higher frequencies
        #myfreqs = np.logspace(np.log10(3), np.log10(150), num=15)
        #myfreqs = [3,4,5,6,7,8]

        #Run TF decomposition and log
        power = time_frequency.tfr_morlet(eeg, freqs=myfreqs, n_cycles=6, return_itc=False, average=False)
        powsa = np.log10(power.data)
        
        # Remove buffer
        buf_samps = int(round(sr*buf/1000.))
        powsa = powsa[:,:,:,buf_samps:-buf_samps] # event x elec x freq x time

        # Average the power across time and frequencies
        powsa_avg = np.nanmean(powsa,axis=4)
        powsa_avg = np.nanmean(powsa_avg,axis=2)
        
        # Z-score the power
        pows_z_sess = zscore(powsa_avg,axis=0)
        pows_z.append(pows_z_sess)

        # Get the correlations of z-scored power with z-scored list performance
        rs_sess = []
        ps_sess = []

        for e in range(pows_z_sess.shape[1]):
            rs_t = []
            ps_t = []
            for t in range(pows_z_sess.shape[2]):
                r, p = spearmanr(pows_z_sess[:,e,t],zscore(perfs_list_sess))
                rs_t.append(r)
                ps_t.append(p)
            rs_sess.append(rs_t)
            ps_sess.append(ps_t)
        
        rs_sess = np.array(rs_sess)
        ps_sess = np.array(ps_sess)

        ps_fdr_sess = (multipletests(ps_sess.flatten(),method='fdr_bh'))[1]

        # Z-transform the correlations
        rs_z_sess = np.arctanh(rs_sess)
        
        # Append to subject level results
        rs_z.append(rs_z_sess)
        ps.append(ps_sess)
        ps_fdr.append(ps_fdr_sess)
        
        # Append session-level result to subject-level result
        #result = 

    # Aggregate results
    result = {'rs_z':rs_z, 'pvals':ps, 'pvals_fdr':ps_fdr, 'pairs':pairs, 'pows_z':pows_z, 'perfs_list':perfs_list, 'perfs_z':perfs_z, 'myfreqs':myfreqs, 'rois':rois, 'tois':tois, 'times':times}
        
        
    return result


def permTest(setA, setB, nperm=250):

    #Function to perform permutation testing on two sets
    nulldist = []
    setAB = np.concatenate([setA,setB])
    
    #Initialize null distribution with true difference
    nulldist.append(np.mean(setA)-np.mean(setB))
    for i in range(nperm-1):
        setAB_p = np.random.permutation(setAB)
        a_p = setAB_p[0:len(setA)]
        b_p = setAB_p[len(setA):]
        nulldist.append(np.mean(a_p)-np.mean(b_p))
        
    pval = np.sum([np.abs(nulldist[0])<=np.abs(d) for d in nulldist])/nperm
    
    return nulldist, pval


def generateList(idx,nrecalls):
    newlist = []
    
    for rec in range(nrecalls):
        np.random.choice(recalls_idx,size=nrecalls,replace=True)

    return

def constructPool(recalls,seglength):
    segments = []
    segrecalls = []
    for ll_ind, ll in enumerate(recalls['list'].unique()):
        reclist = recalls[recalls['list']==ll]
        for i in range(12-seglength+1):
            # record the index out of [length(recalls)]
            seg_idx = (ll_ind)*12 + np.arange(i,i+seglength)
            segments.append(seg_idx)
            seg_rec = np.sum(list(reclist['recalled'])[i:i+seglength])
            segrecalls.append(seg_rec)
    
    return np.array(segments),np.array(segrecalls)
                
        
def generateNewLists(recalls,itemPows,listlength=1,nrecalls=1):
    # Function to generate synthetic lists based on sub-lists of length "listlength", and with constant performance of "nrecalls"
    # And then apply the reshuffling to power arrays.  Assuming just 1 session
    
    # Get the actual number of lists
    lists = recalls['list'].unique()
    lists_idx = np.arange(len(lists))
    recalled = np.array(recalls['recalled'])
    recalls_idx = np.where(recalled==True)[0]

    # Get the list of potential synthetic lists
    newlists = []
    poslists, poslists_recalls = constructPool(recalls,listlength)
    poslists = poslists[[pr == nrecalls for pr in poslists_recalls]]
    
    # Pick random selection of the new possible lists, with replacement
    numsublists = int(np.floor(12/listlength)) # calculate number of sub-lists needed to get to 12
    sublists_idx = np.arange(len(lists)*numsublists) # generate index numbers for each sublist
    newlists_idx = np.arange(len(poslists))
    newlists_sel = np.random.choice(newlists_idx,size=len(lists)*numsublists,replace=True) # get the random selection
    
    # generate the new lists based on the random selection
    newlists = np.array([np.concatenate([poslists[newlists_sel[sel+j]] for j in range(numsublists)]) for sel in sublists_idx[::numsublists]])
    
    # apply the random shuffling to itemPows and listPows arrays
    nlists = int(np.floor(len(itemPows)/12))
    itemPows2 = itemPows[newlists.flatten()]
    listPows2 = np.array([np.nanmean(itemPows[newlists[l]],axis=0) for l in range(nlists)])
    
    return newlists, itemPows2, listPows2


#def restructureArray():
    # For each permutation, generate a new set of lists and compute power
    # arrays to restructure: recalls, itemPows, 'listPows':listPows
    itemPows2 = []
    
    for s in recalls:
        newlists = generateNewLists(recalls[s],listlength=2,nrecalls=1)
        itemPow2 = itemPows[s][newlists.flatten()]
        listPow2 = np.array([np.nanmean(itemPows[s][newlists[l]],axis=0) for l in range(len(listPows[s]))])
        itemPows2.append(itemPow2)
        listPows2.append(listPow2)
        
        
def doListPermTest(recalls,itemPows,listlength,nperm=500): 
    from scipy.stats import ttest_ind, ttest_1samp
    # Generate the new lists for each recall level

    ll = listlength
    nrecs = [0,ll-1]
    ts = []
    for p in range(nperm):
        listPows_all = []
        for nrec in nrecs:
            newlists, _, listPows2 = generateNewLists(recalls,itemPows,listlength=ll,nrecalls=nrec)
            listPows_all.append(listPows2)

        #ttest of listpow and performance
        t, _ = ttest_ind(listPows_all[-1],listPows_all[0])
        ts.append(t)

    ttot = ttest_1samp(ts,0)
    
    return ttot


def logitPerf0(perf):
    perf = np.array(perf)
    prec = perf/12
    prec[prec==1] -= 10e-4
    prec[prec==0] += 10e-4
    return np.log(prec/(1-prec))


def logitPerf0_inv(logitperf):
    prec = np.exp(logitperf)/(1+np.exp(logitperf))
    prec = prec.round(3)
    prec[prec==0.999] += 10e-4
    prec[prec==0.001] -= 10e-4
    return np.round(12*prec,2)


def logitPerf(perf):
    perf = np.array(perf)
    prec = perf/12
    logitperf = np.zeros(perf.shape)
    c = 0.5/12 # formula from https://www.bmj.com/content/352/bmj.i1114.long
    with np.errstate(divide='ignore'):
        logitperf[(prec<1) & (prec>0)] = np.log(prec/(1-prec))[(prec<1) & (prec>0)]
        logitperf[(prec==1) | (prec==0)] = np.log((prec+c)/(1-prec+c))[(prec==1) | (prec==0)]

    return logitperf


def logitPerf_inv(logitperf,roundint=False):
    prec = np.exp(logitperf)/(1+np.exp(logitperf))
    if roundint:
        inv = np.round(prec*12,0)
    else:
        inv = prec*12
    return inv


def normalizePerfs(result,logit=True):
    perfs_all = np.concatenate(result['perfs_list'])
    nsessions = len(result['perfs_list'])
    lists = np.arange(len(np.concatenate(result['lists'])))
    perfs_norm = []
    for list_test in lists:
        listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
        if logit:
            perfs_all_norm, perfmean, perfstd, set_omit = norm_omit_list(logitPerf(perfs_all),listNumsBySess,omit=list_test)
            perfs_test = logitPerf([perfs_all[list_test]])-perfmean # adjust by the mean of the set from which it came
        else:
            perfs_all_norm, perfmean, perfstd, set_omit = norm_omit_list(perfs_all,listNumsBySess,omit=list_test)
            perfs_test = perfs_all[list_test]-perfmean # adjust by the mean of the set from which it came

        perfs_norm.append(perfs_test)

    return perfs_norm


def zscore_omit_item(arr,lengths,omit):
    from scipy.stats import zscore
    
    omit = int(omit)
    nsets = len(lengths)
    sets_idx = [int(np.sum(lengths[0:i])) for i in range(nsets+1)]
        
    # Determine which set omitted index is in
    set_omit = np.arange(nsets)[[sets_idx[i] <= omit < sets_idx[i+1] for i in range(nsets)]][0]
    lengths[set_omit] -= 1 # Reduce length of that set by 1
    sets_idx = [int(np.sum(lengths[0:i])) for i in range(nsets+1)] # Recalculate set indices
    
    # Remove omission from array and norm
    arr = arr[[l != omit for l in range(len(arr))]]
    arr_reshape = arr.reshape(int(arr.shape[0]*arr.shape[1]),arr.shape[2],arr.shape[3])
    arr_znorm = np.concatenate([zscore(arr_reshape[12*sets_idx[i]:12*sets_idx[i+1]],nan_policy='omit') for i in range(nsets)])

    # Calculate the mean and sd for the session from which the list was taken
    mean = np.nanmean(arr[sets_idx[set_omit]:sets_idx[set_omit+1]],axis=(0,1))
    std = np.nanstd(arr[sets_idx[set_omit]:sets_idx[set_omit+1]],axis=(0,1))
    
    return arr_znorm, mean, std, set_omit


def zscore_omit_serialpos(arr,lengths,listlength,omit):
    from scipy.stats import zscore
    
    omit = int(omit)
    nsets = len(lengths)
    sets_idx = [int(np.sum(lengths[0:i])) for i in range(nsets+1)]
        
    # Determine which set omitted index is in
    set_omit = np.arange(nsets)[[sets_idx[i] <= omit < sets_idx[i+1] for i in range(nsets)]][0]
    lengths[set_omit] -= 1 # Reduce length of that set by 1
    sets_idx = [int(np.sum(lengths[0:i])) for i in range(nsets+1)] # Recalculate set indices
    
    # Remove omission from array and norm
    arr = arr[[l != omit for l in range(len(arr))]]
    arr_reshape = arr.reshape(int(arr.shape[0]*arr.shape[1]),arr.shape[2],arr.shape[3])
    arr_znorm = np.concatenate([zscore(arr_reshape[listlength*sets_idx[i]:listlength*sets_idx[i+1]],nan_policy='omit') for i in range(nsets)])

    # Calculate the mean and sd for the session from which the list was taken
    mean = np.nanmean(arr[sets_idx[set_omit]:sets_idx[set_omit+1]],axis=(0,1))
    std = np.nanstd(arr[sets_idx[set_omit]:sets_idx[set_omit+1]],axis=(0,1))
    
    return arr_znorm, mean, std, set_omit


def zscore_omit_list(arr,lengths,omit):
    from scipy.stats import zscore
    
    omit = int(omit)
    nsets = len(lengths)
    sets_idx = [int(np.sum(lengths[0:i])) for i in range(nsets+1)]
        
    # Determine which set omitted index is in
    set_omit = np.arange(nsets)[[sets_idx[i] <= omit < sets_idx[i+1] for i in range(nsets)]][0]
    lengths[set_omit] -= 1 # Reduce length of that set by 1
    sets_idx = [int(np.sum(lengths[0:i])) for i in range(nsets+1)] # Recalculate set indices
    
    # Remove omission from array and norm
    arr = arr[[l != omit for l in range(len(arr))]]
    arr_znorm = np.concatenate([zscore(arr[sets_idx[i]:sets_idx[i+1]],nan_policy='omit') for i in range(nsets)])

    # Calculate the mean and sd for the session from which the list was taken
    mean = np.nanmean(arr[sets_idx[set_omit]:sets_idx[set_omit+1]],axis=0)
    std = np.nanstd(arr[sets_idx[set_omit]:sets_idx[set_omit+1]],axis=0)
    
    return arr_znorm, mean, std, set_omit


def norm_omit_list(arr,lengths,omit):
    from scipy.stats import zscore
    
    omit = int(omit)
    nsets = len(lengths)
    sets_idx = [int(np.sum(lengths[0:i])) for i in range(nsets+1)]
        
    # Determine which set omitted index is in
    set_omit = np.arange(nsets)[[sets_idx[i] <= omit < sets_idx[i+1] for i in range(nsets)]][0]
    lengths[set_omit] -= 1 # Reduce length of that set by 1
    sets_idx = [int(np.sum(lengths[0:i])) for i in range(nsets+1)] # Recalculate set indices
    
    # Remove omission from array and norm
    arr = arr[[l != omit for l in range(len(arr))]]
    arr_norm = np.concatenate([arr[sets_idx[i]:sets_idx[i+1]]-np.mean(arr[sets_idx[i]:sets_idx[i+1]]) for i in range(nsets)])

    # Calculate the mean and sd for the session from which the list was taken
    mean = np.nanmean(arr[sets_idx[set_omit]:sets_idx[set_omit+1]],axis=0)
    std = np.nanstd(arr[sets_idx[set_omit]:sets_idx[set_omit+1]],axis=0)
    
    return arr_norm, mean, std, set_omit


def norm_list(arr,lengths):
    nsets = len(lengths)
    sets_idx = [int(np.sum(lengths[0:i])) for i in range(nsets+1)]

    # Remove omission from array and norm
    arr_norm = np.concatenate([arr[sets_idx[i]:sets_idx[i+1]]-np.mean(arr[sets_idx[i]:sets_idx[i+1]]) for i in range(nsets)])

    return arr_norm


def addAvgRecallabilities(result,subjlist,norm=False):
    rootdir = '/home1/daniel.rubinstein/Projects/listSME/'
    # Get the item recallability of all subjects in subjlist excluding subj (so for purposes of subj)
    recalls_allsub = pd.read_csv(rootdir+'recalls_allsub_eng.csv')
    
    loso_df = recalls_allsub[recalls_allsub['subj'] != result['subject']]
    subjmeans_loso = loso_df.groupby(['item_name','subj']).mean()
    wordmeans_loso = subjmeans_loso.groupby('item_name').mean()
    #result = np.load(resdir+subj+'_'+task+'_itempow.npy',allow_pickle=True).item()
    
    # get recallability of each word they see
    for r, recalls_sub in enumerate(result['recalls']):
        if norm:
            recalls_sub['recallability_norm'] = list(wordmeans_loso.loc[list(recalls_sub['item_name'])]['recalled_norm'])
        else:
            recalls_sub['recallability'] = list(wordmeans_loso.loc[list(recalls_sub['item_name'])]['recalled'])
#         recallabilities1 = list(wordmeans_loso.loc[list(recalls_sub['item_name'])]['recalled'])
#         recallabilities = []
#         for index, row in recalls_sub.iterrows():
#             try:
#                 recallability = wordmeans_loso.loc[row['item_name']]['recalled']
#             except KeyError:
#                 recallability = np.nan
#             recallabilities.append(recallability)
#         #add the recallability to the recalls dataframe
#         recalls_sub['recallability'] = recallabilities1
        
    return result


def recallControlNeuralSignal(result,items_train,list_test,feats_train,feats_test):
    #if recallability_control:
    recallabilities_all = np.concatenate([recs['recallability_norm'] for recs in result['recalls']])
    recallabilities_train = recallabilities_all[items_train].reshape(-1,1)
    reg = LinearRegression().fit(recallabilities_train,feats_train) # regress
    predictions = reg.predict(recallabilities_train) # predict
    feats_train = np.subtract(feats_train, predictions) # residualize

    # apply that model to the test features
    recallabilities_test = recallabilities_all[12*list_test:12*(list_test+1)].reshape(-1,1)
    predictions_test = reg.predict(recallabilities_test)
    feats_test = np.subtract(feats_test,predictions_test)
    
    # do same thing for test features 
    recallabilities_test = recallabilities_all[12*list_test:12*(list_test+1)].reshape(-1,1)
    reg_test = LinearRegression().fit(recallabilities_test,feats_test) # regress
    predictions_test = reg_test.predict(recallabilities_test) # predict
    feats_test = np.subtract(feats_test, predictions_test) # residualize
    
    return feats_train, feats_test


def recallControlItemPerf(result,recalls_all,items_train,list_test):
    from sklearn.linear_model import LinearRegression
    # get average recallabilities for each list
    recallabilities_all = np.concatenate([recs['recallability'] for recs in result['recalls']])
    # split into test and train
    recallabilities_train = recallabilities_all[items_train].reshape(-1,1)
    recallabilities_test = recallabilities_all[12*list_test:12*(list_test+1)].reshape(-1,1)
    recalls_train = recalls_all[items_train]
    recalls_test = recalls_all[12*list_test:12*(list_test+1)]
    # do linear regression on recallability and residualize
    regitem = LinearRegression().fit(recallabilities_train.reshape(-1,1),recalls_train.astype(np.float64))
    recalls_train_rcresid = (recalls_train - regitem.predict(recallabilities_train))
    # apply linear regression parameters of training model to test features as well
    recalls_test_rcresid = (recalls_test.astype(np.float64) - regitem.predict(recallabilities_test))
    # reassign to perfs_all
    recalls_all[items_train] = recalls_train_rcresid
    recalls_all[12*list_test:12*(list_test+1)] = recalls_test_rcresid
    
    return recalls_all, recalls_train_rcresid, recalls_test_rcresid


def recallControlListPerf(result,perfs_all,lists_train,list_test):
    from sklearn.linear_model import LinearRegression
    # get average recallabilities for each list
    listrecallabilities_all = np.concatenate([recs.groupby(['list']).mean()['recallability'] for recs in result['recalls']])
    # split into test and train
    listrecallabilities_train = listrecallabilities_all[lists_train].reshape(-1,1)
    listrecallabilities_test = listrecallabilities_all[list_test].reshape(1,-1)
    perfs_train_reshape = perfs_all[lists_train].reshape(-1,1)
    perfs_test_reshape = perfs_all[list_test].reshape(1,-1)
    # do linear regression on recallability and residualize
    reglist = LinearRegression().fit(listrecallabilities_train,perfs_train_reshape)
    perfs_train_rcresid = (perfs_train_reshape - reglist.predict(listrecallabilities_train))
    # apply linear regression parameters of training model to test features as well
    perfs_test_rcresid = (perfs_test_reshape - reglist.predict(listrecallabilities_test))
    # reassign to perfs_all
    perfs_all[lists_train] = perfs_train_rcresid[:,0]
    perfs_all[list_test] = perfs_test_rcresid[:,0]
    
    return perfs_all, perfs_train_rcresid[:,0], perfs_test_rcresid[:,0]


def classifyPredictItems_lolo(result,list_test,permute=False,balanced_weights=True,recallability_control=False):
    # For each list train classifier on all other lists over all sessions, test on left-out list
    from sklearn.linear_model import LogisticRegression, LinearRegression, Ridge
    from sklearn.metrics import roc_auc_score, roc_curve
    from scipy.stats import zscore
    
    # Concatenate all the lists and performances over all sessions together
    nsessions = len(result['perfs_list'])
    itemPows = np.concatenate([result['itemPows_raw'][s] for s in range(nsessions)])
    
    # Remove electrodes with inf values
    goodelec = np.arange(itemPows.shape[1])[[(e not in np.where(np.isinf(itemPows))[1]) and
                                             (e not in np.where(np.isnan(itemPows))[1]) for e in np.arange(itemPows.shape[1])]]

    itemPows = itemPows[:,goodelec,:]
    
    recalls_all = np.concatenate([recs['recalled'] for recs in result['recalls']])
    recalls_all_norm = np.concatenate([recs['recalled_norm_holdout'] for recs in result['recalls']]) # make sure to use heldout normed to avoid contamination of training to test sets
    
    if permute:
        np.random.shuffle(recalls_all)
    perfs_all = np.concatenate(result['perfs_list'])
    lists = np.arange(len(perfs_all))
    lists_train = lists[[l != list_test for l in lists]]
    items = np.arange(len(itemPows))
    items_train = items[[i not in np.arange(12*list_test,12*(list_test+1)) for i in items]]
    
    # Identify which session the test list came from and normalize items by session
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    itemPows_reshaped = itemPows.reshape(int(itemPows.shape[0]/12),12,itemPows.shape[1],itemPows.shape[2]) # reshape by list
    itemPows_znorm, mean, std, set_omit = zscore_omit_item(itemPows_reshaped,listNumsBySess,omit=list_test) # output has test list omitted
    
    recalls_train = recalls_all[items_train]
    recalls_test = recalls_all[12*list_test:12*(list_test+1)]

    if recallability_control:
        recalls_all, recalls_train, recalls_test = recallControlItemPerf(result,recalls_all,items_train,list_test)

    pow_train = itemPows_znorm
    pow_test = itemPows_reshaped[list_test]
    pow_test_norm = np.subtract(pow_test,mean)/std
    
    feats_train = pow_train.reshape(len(recalls_train), -1)
    feats_test = pow_test_norm.reshape(12, -1)
        
    # Logistic Regression Classifier
    C = 0.0007    

    if not recallability_control:
        if balanced_weights:
            lr_classifier = LogisticRegression(C=C, penalty='l2', solver='liblinear', multi_class='ovr', n_jobs=1,
                                          class_weight='balanced')

        else:
            lr_classifier = LogisticRegression(C=C, penalty='l2', solver='liblinear', multi_class='ovr', n_jobs=1)
        lr_classifier.fit(feats_train, recalls_train)

        # Predict recalls of items from left-out list
        preds = lr_classifier.predict(feats_test)

        # Calculate ROC and AUC
        preds_probs = lr_classifier.predict_proba(feats_test)[:,1]
        #auc = roc_auc_score(recalls_test, preds_probs)
        #roc = roc_curve(recalls_test, preds_probs)
    else:
        # then do linear regression because recall now continuous variable
        alpha = 1/(2*C)
        ridge_classifier = Ridge(alpha=alpha)
        ridge_classifier.fit(feats_train, recalls_train)
        # Predict recalls of items from left-out list
        preds_probs = ridge_classifier.predict(feats_test)
        preds = preds_probs # actual predictions can't be done because not a logistic regression

    return recalls_test, preds, preds_probs


def classifyPredictItems_lolo_serialpos(result,list_test,serialpos,balanced_weights=True):
    # For each list train classifier on all other lists over all sessions, test on left-out list
    # Further, narrow by set of serial positions
    from sklearn.linear_model import LogisticRegression
    from sklearn.metrics import roc_auc_score, roc_curve
    from scipy.stats import zscore
    
    nsessions = len(result['perfs_list'])
    listlength = len(serialpos)
    
    # Filter by serial position
    sps = np.concatenate([list(result['recalls'][s]['serialpos']) for s in range(nsessions)])
    for s in range(nsessions):
        sp_filter = list(result['recalls'][s]['serialpos'].isin(serialpos))
        for key in ['recalls','itemPows_raw','itemPows']:
            result[key][s] = result[key][s][sp_filter]
    
    # Concatenate all the lists and performances over all sessions together
    itemPows = np.concatenate([result['itemPows_raw'][s] for s in range(nsessions)])
        
    # Remove electrodes with inf values
    goodelec = np.arange(itemPows.shape[1])[[(e not in np.where(np.isinf(itemPows))[1]) and
                                             (e not in np.where(np.isnan(itemPows))[1]) for e in np.arange(itemPows.shape[1])]]

    itemPows = itemPows[:,goodelec,:]
    
    recalls_all = np.concatenate([result['recalls'][s]['recalled'] for s in range(nsessions)])
    perfs_all = np.concatenate(result['perfs_list'])
    lists = np.arange(len(perfs_all))
    lists_train = lists[[l != list_test for l in lists]]
    items = np.arange(len(itemPows))
    items_train = items[[i not in np.arange(listlength*list_test,listlength*(list_test+1)) for i in items]]
    
    # Identify which session the test list came from and normalize items by session
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    itemPows_reshaped = itemPows.reshape(int(itemPows.shape[0]/listlength),listlength,itemPows.shape[1],itemPows.shape[2]) # reshape by list
    itemPows_znorm, mean, std, set_omit = zscore_omit_serialpos(itemPows_reshaped,listNumsBySess,listlength,omit=list_test) # output has test list omitted
    
    recalls_train = recalls_all[items_train]
    recalls_test = recalls_all[listlength*list_test:listlength*(list_test+1)]

    pow_train = itemPows_znorm
    pow_test = itemPows_reshaped[list_test]
    pow_test_norm = np.subtract(pow_test,mean)/std

    feats_train = pow_train.reshape(len(recalls_train), -1)
    feats_test = pow_test_norm.reshape(listlength, -1)
        
    # Logistic Regression Classifier
    C = 0.0007
    if balanced_weights:
        lr_classifier = LogisticRegression(C=C, penalty='l2', solver='liblinear', multi_class='ovr', n_jobs=1,
                                      class_weight='balanced')
    else:
        lr_classifier = LogisticRegression(C=C, penalty='l2', solver='liblinear', multi_class='ovr', n_jobs=1)
    lr_classifier.fit(feats_train, recalls_train)
    
    # Predict recalls of items from left-out list
    preds = lr_classifier.predict(feats_test)
    
    # Calculate ROC and AUC
    preds_probs = lr_classifier.predict_proba(feats_test)[:,1]

    return recalls_test, preds, preds_probs


def classifyPredictLists_lolo(result,list_test,permute=False,perfnorm=True,recallability_control=False):
    # For each list train classifier on all other lists over all sessions, test on left-out list
    from sklearn.linear_model import LogisticRegression, Ridge, LinearRegression
    from scipy.stats import zscore
    
    # Concatenate all the lists and performances over all sessions together
    nsessions = len(result['perfs_list'])
    listPows = np.concatenate([result['listPows_raw'][s] for s in range(nsessions)])
    
    # Remove electrodes with inf and nan values
    goodelec = np.arange(listPows.shape[1])[[(e not in np.where(np.isinf(listPows))[1]) and
                                             (e not in np.where(np.isnan(listPows))[1]) for e in np.arange(listPows.shape[1])]]

    listPows = listPows[:,goodelec,:]

    lists = np.arange(len(listPows))
    lists_train = lists[[l != list_test for l in lists]]
    
    perfs_all = np.concatenate(result['perfs_list']).astype(np.float64)
    
    if permute:
        np.random.shuffle(perfs_all)
    #perfs_all_norm = np.concatenate([perfs-np.mean(perfs) for perfs in result['perfs_list']])
    
    # Identify which session the test list came from and normalize training lists by session
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    listPows_znorm, mean, std, set_omit = zscore_omit_list(listPows,listNumsBySess,omit=list_test) # this already has the test list omitted
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]

    perfs_all_norm, perfmean, perfstd, _ = norm_omit_list(logitPerf(perfs_all),listNumsBySess,omit=list_test)
    
    perfs_test = perfs_all[list_test]
    
    if perfnorm:
        perfs_train = perfs_all_norm # this is already logit-transformed
        perfs_test = logitPerf([perfs_test])-perfmean # adjust by the mean from the set which it came from
    else:
        perfs_train = perfs_all[lists_train] # this will be logit-transformed later
        perfs_test = logitPerf([perfs_test])

    # adjust performance by recallability
    if recallability_control:
        perfs_all[lists_train] = perfs_train
        perfs_all[list_test] = perfs_test
        perfs_all, perfs_train, perfs_test = recallControlListPerf(result,perfs_all,lists_train,list_test)
    
    pow_train = listPows_znorm
    pow_test = listPows[list_test]
    pow_test_norm = np.subtract(pow_test,mean)/std

    feats_train = pow_train.reshape(len(perfs_train), -1)
    feats_test = pow_test_norm.reshape(1, -1)
    
    # Ridge Regression Classifier
    C = 0.0007
    alpha = 1/(2*C)
    ridge_classifier = Ridge(alpha=alpha)
    if perfnorm or recallability_control:
        ridge_classifier.fit(feats_train, perfs_train)
    else:
        ridge_classifier.fit(feats_train, logitPerf(perfs_train))

    # Predict performance of left-out list
    preds_ridge = ridge_classifier.predict(feats_test)

    return perfs_test, preds_ridge


def classifyPredictLists_reduced_lolo(result,list_test,nsessions_to_include=0,permute=False,perfnorm=False):
    # For each list train classifier on all other lists over all sessions, test on left-out list
    from sklearn.linear_model import LogisticRegression, Ridge
    from scipy.stats import zscore
    
    nsessions = len(result['perfs_list'])
    
    if nsessions_to_include > 0:
        # pick random sessions to include
        includes = np.random.choice(np.arange(nsessions),nsessions_to_include)
        for key in 'listPows_raw','perfs_list':
            result[key] = result[key][includes]
        nsessions = len(result['perfs_list'])
    
    # Concatenate all the lists and performances over all sessions together
    listPows = np.concatenate([result['listPows_raw'][s] for s in range(nsessions)])
    
    # Remove electrodes with inf and nan values
    goodelec = np.arange(listPows.shape[1])[[(e not in np.where(np.isinf(listPows))[1]) and
                                             (e not in np.where(np.isnan(listPows))[1]) for e in np.arange(listPows.shape[1])]]

    listPows = listPows[:,goodelec,:]

    perfs_all = np.concatenate(result['perfs_list'])
    
    lists = np.arange(len(listPows))
    lists_train = lists[[l != list_test for l in lists]]
    
    # Identify which session the test list came from and normalize training lists by session
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    listPows_znorm, mean, std, set_omit = zscore_omit_list(listPows,listNumsBySess,omit=list_test) # this already has the test list omitted
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    perfs_all_norm, perfmean, perfstd, _ = norm_omit_list(logitPerf(perfs_all),listNumsBySess,omit=list_test)
    
    perfs_test = perfs_all[list_test]
    
    if perfnorm:
        perfs_train = perfs_all_norm # this is already logit transformed
        perfs_test = logitPerf([perfs_test])-perfmean # subtract mean of set from which it came
    else:
        perfs_train = perfs_all[lists_train] # this will be logit transformed later
        perfs_test = logitPerf([perfs_test])
    
    if permute:
        np.random.shuffle(perfs_train)

    pow_train = listPows_znorm
    pow_test = listPows[list_test]
    pow_test_norm = np.subtract(pow_test,mean)/std

    feats_train = pow_train.reshape(len(perfs_train), -1)
    feats_test = pow_test_norm.reshape(1, -1)
        
    # Ridge Regression Classifier
    C = 0.0007
    alpha = 1/(2*C)
    ridge_classifier = Ridge(alpha=alpha)
    if perfnorm:
        ridge_classifier.fit(feats_train, perfs_train)
    else:
        ridge_classifier.fit(feats_train, logitPerf(perfs_train))

    # Predict performance of left-out list
    preds_ridge = ridge_classifier.predict(feats_test)

    return perfs_test, preds_ridge


def classifyPredictSlidingWindowByTimeFreq_lolo(result_wholeList,result_avg,list_test,permute=False,perfnorm=False,shift=0,winsize=1000,avglistpow=True):
    # For each list train classifier on all other lists over all sessions, test on left-out list
    from sklearn.linear_model import LogisticRegression, Ridge
    from scipy.stats import zscore
    
    # Concatenate all the lists and performances over all sessions together
    nsessions = len(result_wholeList['perfs_list'])
    wholeListPows = np.concatenate([result_wholeList['wholeListPows_raw'][s] for s in range(nsessions)])
    if avglistpow:
        listPows = np.concatenate([res for res in result_avg['avgListPows_raw']]) # use average power of the exact encoding list time windows
    else:
        listPows = np.nanmean(wholeListPows,3) # Instead of loading listpows from 300-1600, use average from currently loaded "wholeListPows"
    #listPows = np.concatenate([result['listPows_raw'][s] for s in range(nsessions)])
    
    # Remove electrodes with inf and nan values
    goodelec = np.arange(wholeListPows.shape[1])[[(e not in np.where(np.isinf(wholeListPows))[1]) and
                                             (e not in np.where(np.isnan(wholeListPows))[1]) for e in np.arange(wholeListPows.shape[1])]]
    #goodelec2 = np.arange(listPows.shape[1])[[(e not in np.where(np.isinf(listPows))[1]) and
    #                                         (e not in np.where(np.isnan(listPows))[1]) for e in np.arange(listPows.shape[1])]]

    #goodelec = np.array(list(set(goodelec1).intersection(set(goodelec2))))
    wholeListPows = wholeListPows[:,goodelec,:,:]
    listPows = listPows[:,goodelec,:]
    #perfs_all = np.concatenate(result['perfs_list'])
    # rotational shift of the performances by shift number
    perfs_all = np.concatenate([vecshift(perfs,shift) for perfs in result_wholeList['perfs_list']])
    
    lists = np.arange(len(wholeListPows))
    lists_train = lists[[l != list_test for l in lists]]
    
    # Identify which session the test list came from and normalize training lists by session
    listNumsBySess = [len(result_wholeList['perfs_list'][s]) for s in range(nsessions)]
    listPows_znorm, mean, std, set_omit = zscore_omit_list(listPows,listNumsBySess,omit=list_test) # this already has the test list omitted   
    listNumsBySess = [len(result_wholeList['perfs_list'][s]) for s in range(nsessions)]
    perfs_all_norm, perfmean, perfstd, _ = norm_omit_list(logitPerf(perfs_all),listNumsBySess,omit=list_test)

    perfs_test = perfs_all[list_test]
    
    # Normalize the listperfs by the average
    # Extend through time the mean and std of lists
    numwins = wholeListPows.shape[-1]
    mean_time = np.repeat(np.expand_dims(mean,2),numwins,axis=2)
    std_time = np.repeat(np.expand_dims(std,2),numwins,axis=2)
    # Remove test list
    wholeListPows_znorm = wholeListPows[lists_train]
    wholeListPows_znorm = np.subtract(wholeListPows_znorm,mean_time)/std_time

    if perfnorm:
        perfs_train = perfs_all_norm # this is already logit-transformed
        perfs_test = logitPerf([perfs_test])-perfmean # adjust by the mean from the set which it came from
    else:
        perfs_train = perfs_all[lists_train] # this will be logit-transformed later
        perfs_test = logitPerf([perfs_test])
    if permute:
        np.random.shuffle(perfs_train)
    
    # Arrange frequencies and times
    myfreqs = result_wholeList['myfreqs']
    if winsize==1000:
        timewins = result_wholeList['times_final'][0:311]
    elif winsize==3000:
        timewins = result_wholeList['times_final'][0:97]
    elif winsize==6000:
        timewins = result_wholeList['times_final'][0:44]
    elif winsize==10000:
        timewins = result_wholeList['times_final'][0:23]
    elif winsize==15000:
        timewins = result_wholeList['times_final'][0:12]
        
    # Start sub-loop for specific time-frequency cell
    preds = np.zeros((len(myfreqs),len(timewins)))
    preds[:] = np.nan

    for f,freq in enumerate(myfreqs):
        for t,timewin in enumerate(timewins):
            subListPows_znorm = wholeListPows_znorm[:,:,f,t]

            pow_train = subListPows_znorm
            pow_test = wholeListPows[list_test,:,f,t]
            pow_test_norm = np.subtract(pow_test,mean_time[:,f,t])/std_time[:,f,t]

            feats_train = pow_train.reshape(len(perfs_train), -1)
            feats_test = pow_test_norm.reshape(1, -1)
                
            # Ridge Regression Classifier
            C = 0.0007
            alpha = 1/(2*C)
            ridge_classifier = Ridge(alpha=alpha)
            if perfnorm:
                ridge_classifier.fit(feats_train, perfs_train)
            else:
                ridge_classifier.fit(feats_train, logitPerf(perfs_train))

            # Predict performance of left-out list
            preds_ridge = ridge_classifier.predict(feats_test)

            # Attach to original matrix of predictions
            preds[f,t] = preds_ridge

    return perfs_test, preds


def classifyPredictSlidingWindowByTime_lolo(result_wholeList,result_avg,list_test,permute=False,perfnorm=False,shift=0,winsize=1000,avglistpow=True):
    # For each list train classifier on all other lists over all sessions, test on left-out list
    from sklearn.linear_model import LogisticRegression, Ridge
    from scipy.stats import zscore
    
    # Concatenate all the lists and performances over all sessions together
    nsessions = len(result_wholeList['perfs_list'])
    wholeListPows = np.concatenate([result_wholeList['wholeListPows_raw'][s] for s in range(nsessions)])
    if avglistpow:
        listPows = np.concatenate([res for res in result_avg['avgListPows_raw']]) # use average power of the exact encoding list time windows
    else:
        listPows = np.nanmean(wholeListPows,3) # Instead of loading listpows from 300-1600, use average from currently loaded "wholeListPows"
    #listPows = np.concatenate([result['listPows_raw'][s] for s in range(nsessions)])
    
    # Remove electrodes with inf and nan values
    goodelec = np.arange(wholeListPows.shape[1])[[(e not in np.where(np.isinf(wholeListPows))[1]) and
                                             (e not in np.where(np.isnan(wholeListPows))[1]) for e in np.arange(wholeListPows.shape[1])]]
    #goodelec2 = np.arange(listPows.shape[1])[[(e not in np.where(np.isinf(listPows))[1]) and
    #                                         (e not in np.where(np.isnan(listPows))[1]) for e in np.arange(listPows.shape[1])]]

    #goodelec = np.array(list(set(goodelec1).intersection(set(goodelec2))))
    wholeListPows = wholeListPows[:,goodelec,:,:]
    listPows = listPows[:,goodelec,:]
    #perfs_all = np.concatenate(result['perfs_list'])
    # rotational shift of the performances by shift number
    perfs_all = np.concatenate([vecshift(perfs,shift) for perfs in result_wholeList['perfs_list']])
    
    lists = np.arange(len(wholeListPows))
    lists_train = lists[[l != list_test for l in lists]]
    
    # Identify which session the test list came from and normalize training lists by session
    listNumsBySess = [len(result_wholeList['perfs_list'][s]) for s in range(nsessions)]
    listPows_znorm, mean, std, set_omit = zscore_omit_list(listPows,listNumsBySess,omit=list_test) # this already has the test list omitted  
    listNumsBySess = [len(result_wholeList['perfs_list'][s]) for s in range(nsessions)]
    perfs_all_norm, perfmean, perfstd, _ = norm_omit_list(logitPerf(perfs_all),listNumsBySess,omit=list_test)

    perfs_test = perfs_all[list_test]
    
    # Normalize the listperfs by the average
    # Extend through time the mean and std of lists
    numwins = wholeListPows.shape[-1]
    mean_time = np.repeat(np.expand_dims(mean,2),numwins,axis=2)
    std_time = np.repeat(np.expand_dims(std,2),numwins,axis=2)
    # Remove test list
    wholeListPows_znorm = wholeListPows[lists_train]
    wholeListPows_znorm = np.subtract(wholeListPows_znorm,mean_time)/std_time

    if perfnorm:
        perfs_train = perfs_all_norm # this is already logit-transformed
        perfs_test = logitPerf([perfs_test])-perfmean # adjust by the mean from the set which it came from
    else:
        perfs_train = perfs_all[lists_train] # this will be logit-transformed later
        perfs_test = logitPerf([perfs_test])
    if permute:
        np.random.shuffle(perfs_train)
    
    # Arrange frequencies and times
    myfreqs = result_wholeList['myfreqs']
    if winsize==1000:
        timewins = result_wholeList['times_final'][0:311]
    elif winsize==3000:
        timewins = result_wholeList['times_final'][0:97]
    elif winsize==6000:
        timewins = result_wholeList['times_final'][0:44]
    elif winsize==10000:
        timewins = result_wholeList['times_final'][0:23]
    elif winsize==15000:
        timewins = result_wholeList['times_final'][0:12]


    # Start sub-loop for specific time-frequency cell
    preds = np.zeros((len(timewins)))
    preds[:] = np.nan

    for t,timewin in enumerate(timewins):
        subListPows_znorm = wholeListPows_znorm[:,:,:,t]

        pow_train = subListPows_znorm
        pow_test = wholeListPows[list_test,:,:,t]
        pow_test_norm = np.subtract(pow_test,mean_time[:,:,t])/std_time[:,:,t]

        feats_train = pow_train.reshape(len(perfs_train), -1)
        feats_test = pow_test_norm.reshape(1, -1)

        # Ridge Regression Classifier
        C = 0.0007
        alpha = 1/(2*C)
        ridge_classifier = Ridge(alpha=alpha)
        if perfnorm:
            ridge_classifier.fit(feats_train, perfs_train)
        else:
            ridge_classifier.fit(feats_train, logitPerf(perfs_train))

        # Predict performance of left-out list
        preds_ridge = ridge_classifier.predict(feats_test)

        # Attach to original matrix of predictions
        preds[t] = preds_ridge

    return perfs_test, preds


def classifyPredictSlidingWindow_item_lolo(result,result_wholeList,list_test,freqs='all'):
    # For each list train classifier on all other lists over all sessions, test on left-out list
    from sklearn.linear_model import LogisticRegression
    
    # Concatenate all the lists and performances over all sessions together
    nsessions = len(result['perfs_list'])
    listnums = np.concatenate(result['lists'])
    itemPows = np.concatenate([result['itemPows_raw'][s] for s in range(nsessions)])
    wholeListPows = np.concatenate([result_wholeList['wholeListPows_raw'][s] for s in range(nsessions)]) # concatentating the lists assumes equal number of time points per list/session
    perfs_all = np.concatenate(result['perfs_list'])
    
    # Remove electrodes with inf values
    goodelec1 = np.arange(itemPows.shape[1])[[(e not in np.where(np.isinf(itemPows))[1]) and
                                              (e not in np.where(np.isnan(itemPows))[1]) for e in np.arange(itemPows.shape[1])]]
    goodelec2 = np.arange(wholeListPows.shape[1])[[(e not in np.where(np.isinf(wholeListPows))[1]) and
                                                   (e not in np.where(np.isnan(wholeListPows))[1]) for e in np.arange(wholeListPows.shape[1])]]
    goodelec = np.array(list(set(goodelec1).intersection(set(goodelec2))))
    
    itemPows = itemPows[:,goodelec,:]
    wholeListPows = wholeListPows[:,goodelec,:,:]
    
    # Make sure frequencies align
    if freqs == 'trunc':
        itemPows = itemPows[:,:,-7:]
        wholeListPows = wholeListPows[:,:,-7:,:]
    elif freqs == 'highg':
        listPows = listPows[:,:,-3:]
        wholeListPows = wholeListPows[:,:,-3:,:]
    
    # Arrange lists and items
    lists = np.arange(len(perfs_all))
    lists_train = lists[[l != list_test for l in lists]]
    items = np.arange(len(itemPows))
    items_train = items[[i not in np.arange(12*list_test,12*(list_test+1)) for i in items]]
    
    # Identify which session the test list came from and normalize training items by session
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    itemPows_reshaped = itemPows.reshape(int(itemPows.shape[0]/12),12,itemPows.shape[1],itemPows.shape[2]) # reshape by list
    itemPows_znorm, mean, std, set_omit = zscore_omit_item(itemPows_reshaped,listNumsBySess,omit=list_test) # output has test list omitted
    pow_train = itemPows_znorm
    
    # Arrange training and test performances
    recalls_all = np.concatenate([result['recalls'][s]['recalled'] for s in range(nsessions)])
    recalls_train = recalls_all[items_train]
    listnum_test = listnums[list_test]
    recalls_test = result_wholeList['recalls'][set_omit][result_wholeList['recalls'][set_omit]['list']==listnum_test]
    recalled_test = list(recalls_test['recalled'])
    times_words = list(recalls_test['mstime_rel'])
    times = result_wholeList['times_final']
    
    # apply norm to sliding windowed power
    pow_test = wholeListPows[list_test]
    numwins = wholeListPows.shape[-1]
    
    # expand mean and std over time
    mean = np.repeat(np.expand_dims(mean,2),numwins,axis=2)
    std = np.repeat(np.expand_dims(std,2),numwins,axis=2)
    pow_test_norm = np.subtract(pow_test,mean)/std

    feats_train = pow_train.reshape(len(recalls_train), -1)
    # train classifier first
    C = 0.0007
    lr_classifier = LogisticRegression(C=C, penalty='l2', solver='liblinear', multi_class='ovr', n_jobs=1, class_weight='balanced')
    lr_classifier.fit(feats_train, recalls_train)
    
    # test classifier on all timewins
    feats_test = np.array([pow_test_norm[:,:,timewin].reshape(1, -1) for timewin in range(numwins)]).squeeze(1)
    preds_prob = lr_classifier.predict_proba(feats_test)[:,1]

    return recalled_test, preds_prob, times, times_words


def classifyPredictSlidingWindow_item_peristim_lolo(result,result_peristim,list_test,freqs='all'):
    # For each list train classifier on all other lists over all sessions, test on left-out list
    from sklearn.linear_model import LogisticRegression
    
    # Concatenate all the lists and performances over all sessions together
    nsessions = len(result['perfs_list'])
    listnums = np.concatenate(result['lists'])
    itemPows = np.concatenate([result['itemPows_raw'][s] for s in range(nsessions)])
    periStimPows = np.concatenate([result_peristim['itemPows_raw'][s] for s in range(nsessions)]) # concatentating the lists assumes equal number of time points per list/session
    perfs_all = np.concatenate(result['perfs_list'])
    
    # Remove electrodes with inf values
    goodelec1 = np.arange(itemPows.shape[1])[[(e not in np.where(np.isinf(itemPows))[1]) and
                                              (e not in np.where(np.isnan(itemPows))[1]) for e in np.arange(itemPows.shape[1])]]
    goodelec2 = np.arange(periStimPows.shape[1])[[(e not in np.where(np.isinf(periStimPows))[1]) and
                                                  (e not in np.where(np.isnan(periStimPows))[1]) for e in np.arange(periStimPows.shape[1])]]
    goodelec = np.array(list(set(goodelec1).intersection(set(goodelec2))))
    
    itemPows = itemPows[:,goodelec,:]
    periStimPows = periStimPows[:,goodelec,:,:]
    
    # Make sure frequencies align
    if freqs == 'trunc':
        listPows = listPows[:,:,-7:]
        periStimPows = periStimPows[:,:,-7:,:]
    
    # Reshape periStim array to be list x item x elec x freq x time
    periStimPows_reshaped = periStimPows.reshape(int(periStimPows.shape[0]/12),12,periStimPows.shape[1],periStimPows.shape[2],periStimPows.shape[3])

    # Arrange lists
    lists = np.arange(len(perfs_all))
    lists_train = lists[[l != list_test for l in lists]]
    items = np.arange(len(itemPows))
    items_train = items[[i not in np.arange(12*list_test,12*(list_test+1)) for i in items]]
    
    # Identify which session the test list came from and normalize training lists by session
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    itemPows_reshaped = itemPows.reshape(int(itemPows.shape[0]/12),12,itemPows.shape[1],itemPows.shape[2]) # reshape by list
    itemPows_znorm, mean, std, set_omit = zscore_omit_item(itemPows_reshaped,listNumsBySess,omit=list_test) # this already has the test list omitted
    pow_train = itemPows_znorm
    
    # Arrange training and test performances
    recalls_all = np.concatenate([result['recalls'][s]['recalled'] for s in range(nsessions)])
    recalls_train = recalls_all[items_train]
    listnum_test = listnums[list_test]
    recalls_test = result_peristim['recalls'][set_omit][result_peristim['recalls'][set_omit]['list']==listnum_test]
    recalled_test = list(recalls_test['recalled'])
    times = result_peristim['times_final']
    
    # apply norm to sliding windowed power
    pow_test = periStimPows_reshaped[list_test]
    numwins = periStimPows_reshaped.shape[-1]
    # expand mean and std over time and items
    mean = np.repeat(np.expand_dims(mean,2),numwins,axis=2) # over time
    mean = np.repeat(np.expand_dims(mean,0),12,axis=0) # over items
    std = np.repeat(np.expand_dims(std,2),numwins,axis=2)
    std = np.repeat(np.expand_dims(std,0),12,axis=0)
    pow_test_norm = np.subtract(pow_test,mean)/std

    feats_train = pow_train.reshape(len(recalls_train), -1)
    # train classifier first
    C = 0.0007
    lr_classifier = LogisticRegression(C=C, penalty='l2', solver='liblinear', multi_class='ovr', n_jobs=1, class_weight='balanced')
    lr_classifier.fit(feats_train, recalls_train)
    
    preds_prob = [[] for i in range(12)]
    # test classifier on loop for each item
    for it in range(12):
        feats_test = np.array([pow_test_norm[it,:,:,timewin].reshape(1, -1) for timewin in range(numwins)]).squeeze(1)
        preds_prob[it] = lr_classifier.predict_proba(feats_test)[:,1]


    # then average over items in the list
    preds_mean = np.nanmean(np.array(preds_prob),0)
    
    return recalled_test, preds_mean, times


def classifyPredictSlidingWindow_lolo(result,result_wholeList,result_avg,list_test,freqs='all',perfnorm=False,wholelistavg=False):
    # For each list train classifier on all other lists over all sessions, test on left-out list
    # normType 0: normalize test features based on test 
    from sklearn.linear_model import Ridge
    
    # Concatenate all the lists and performances over all sessions together
    nsessions = len(result['perfs_list'])
    listnums = np.concatenate(result['lists'])
    wholeListPows = np.concatenate([res for res in result_wholeList['wholeListPows_raw']]) # concatentating the lists assumes equal number of time points per list/session
    if wholelistavg:
        #listPows = np.nanmean(wholeListPows,3)
        listPows = np.concatenate([res for res in result_avg['avgListPows_raw']])
    else:
        listPows = np.concatenate([res for res in result['listPows_raw']])
    perfs_all = np.concatenate(result['perfs_list'])
    
    # Remove electrodes with inf values
    goodelec1 = np.arange(listPows.shape[1])[[(e not in np.where(np.isinf(listPows))[1]) and
                                              (e not in np.where(np.isnan(listPows))[1]) for e in np.arange(listPows.shape[1])]]
    goodelec2 = np.arange(wholeListPows.shape[1])[[(e not in np.where(np.isinf(wholeListPows))[1]) and
                                                   (e not in np.where(np.isnan(wholeListPows))[1]) for e in np.arange(wholeListPows.shape[1])]]
    goodelec = np.array(list(set(goodelec1).intersection(set(goodelec2))))
    listPows = listPows[:,goodelec,:]
    wholeListPows = wholeListPows[:,goodelec,:,:]
    
    # Make sure frequencies align
    if freqs == 'trunc':
        listPows = listPows[:,:,-7:]
        wholeListPows = wholeListPows[:,:,-7:,:]
    elif freqs == 'highg':
        listPows = listPows[:,:,-3:]
        wholeListPows = wholeListPows[:,:,-3:,:]
    
    # Arrange lists
    lists = np.arange(len(listPows))
    lists_train = lists[[l != list_test for l in lists]]
    
    # Identify which session the test list came from and normalize training lists by session
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    listPows_znorm, mean, std, set_omit = zscore_omit_list(listPows,listNumsBySess,omit=list_test) # this already has the test list omitted
    #wholeListPows_znorm, mean, std, set_omit = zscore_omit_list(np.nanmean(wholeListPows,3),listNumsBySess,omit=list_test)
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    perfs_all_norm, perfmean, perfstd, _ = norm_omit_list(logitPerf(perfs_all),listNumsBySess,omit=list_test)
    
    # Arrange training and test performances
    perfs_test = perfs_all[list_test]
    
    if perfnorm:
        perfs_train = perfs_all_norm # this is already logit transformed
        perfs_test = logitPerf([perfs_test])-perfmean # adjust by the mean from the set which it came from
    else:
        perfs_train = perfs_all[lists_train] # this will be logit transformed later
        perfs_test = logitPerf([perfs_test])
        
    listnum_test = listnums[list_test]
    recalls_test = result_wholeList['recalls'][set_omit][result_wholeList['recalls'][set_omit]['list']==listnum_test]
    recalled_test = list(recalls_test['recalled'])
    times_words = list(recalls_test['mstime_rel'])
    times = result_wholeList['times_final']
    
    # apply norm to sliding windowed power
    pow_train = listPows_znorm
    pow_test = wholeListPows[list_test]
    numwins = wholeListPows.shape[-1]
    
    # expand mean and std over time
    mean = np.repeat(np.expand_dims(mean,2),numwins,axis=2)
    std = np.repeat(np.expand_dims(std,2),numwins,axis=2)
    pow_test_norm = np.subtract(pow_test,mean)/std

    feats_train = pow_train.reshape(len(perfs_train), -1)
    # train classifier first
    C = 0.0007
    alpha = 1/(2*C)
    ridge_classifier = Ridge(alpha=alpha)
    if perfnorm:
        ridge_classifier.fit(feats_train, perfs_train)
    else:
        ridge_classifier.fit(feats_train, logitPerf(perfs_train))
    
    preds = []
    # test classifier on loop
    for timewin in range(numwins):
        feats_test = pow_test_norm[:,:,timewin].reshape(1, -1)

        # Predict performance of left-out list
        preds_ridge = ridge_classifier.predict(feats_test)
        preds.extend(preds_ridge)
        
        #preds = np.array([ridge_classifier.predict(pow_test_norm[:,:,timewin].reshape(1, -1)) for timewin in range(numwins)]).flatten()

    return perfs_test, preds, times, times_words


def classifyPredictSlidingWindow_lolo_old(result,result_wholeList,result_avg,list_test,freqs='all',perfnorm=False,wholelistavg=False):
    # For each list train classifier on all other lists over all sessions, test on left-out list
    # normType 0: normalize test features based on test 
    from sklearn.linear_model import Ridge
    
    # Concatenate all the lists and performances over all sessions together
    nsessions = len(result['perfs_list'])
    listnums = np.concatenate(result['lists'])
    wholeListPows = np.concatenate([res for res in result_wholeList['wholeListPows_raw']]) # concatentating the lists assumes equal number of time points per list/session
    if wholelistavg:
        listPows = np.nanmean(wholeListPows,3)
        #listPows = np.concatenate([res for res in result_avg['avgListPows_raw']])
    else:
        listPows = np.concatenate([res for res in result['listPows_raw']])
    perfs_all = np.concatenate(result['perfs_list'])
    
    # Remove electrodes with inf values
    goodelec1 = np.arange(listPows.shape[1])[[(e not in np.where(np.isinf(listPows))[1]) and
                                              (e not in np.where(np.isnan(listPows))[1]) for e in np.arange(listPows.shape[1])]]
    goodelec2 = np.arange(wholeListPows.shape[1])[[(e not in np.where(np.isinf(wholeListPows))[1]) and
                                                   (e not in np.where(np.isnan(wholeListPows))[1]) for e in np.arange(wholeListPows.shape[1])]]
    goodelec = np.array(list(set(goodelec1).intersection(set(goodelec2))))
    listPows = listPows[:,goodelec,:]
    wholeListPows = wholeListPows[:,goodelec,:,:]
    
    # Make sure frequencies align
    if freqs == 'trunc':
        listPows = listPows[:,:,-7:]
        wholeListPows = wholeListPows[:,:,-7:,:]
    elif freqs == 'highg':
        listPows = listPows[:,:,-3:]
        wholeListPows = wholeListPows[:,:,-3:,:]
    
    # Arrange lists
    lists = np.arange(len(listPows))
    lists_train = lists[[l != list_test for l in lists]]
    
    # Identify which session the test list came from and normalize training lists by session
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    listPows_znorm, mean, std, set_omit = zscore_omit_list(listPows,listNumsBySess,omit=list_test) # this already has the test list omitted
    #wholeListPows_znorm, mean, std, set_omit = zscore_omit_list(np.nanmean(wholeListPows,3),listNumsBySess,omit=list_test)
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    perfs_all_norm, perfmean, perfstd, _ = norm_omit_list(logitPerf(perfs_all),listNumsBySess,omit=list_test)
    
    # Arrange training and test performances
    perfs_test = perfs_all[list_test]
    
    if perfnorm:
        perfs_train = perfs_all_norm # this is already logit transformed
        perfs_test = logitPerf([perfs_test])-perfmean # adjust by the mean from the set which it came from
    else:
        perfs_train = perfs_all[lists_train] # this will be logit transformed later
        perfs_test = logitPerf([perfs_test])
        
    listnum_test = listnums[list_test]
    recalls_test = result_wholeList['recalls'][set_omit][result_wholeList['recalls'][set_omit]['list']==listnum_test]
    recalled_test = list(recalls_test['recalled'])
    times_words = list(recalls_test['mstime_rel'])
    times = result_wholeList['times_final']
    
    # apply norm to sliding windowed power
    pow_train = listPows_znorm
    pow_test = wholeListPows[list_test]
    numwins = wholeListPows.shape[-1]
    
    # expand mean and std over time
    mean = np.repeat(np.expand_dims(mean,2),numwins,axis=2)
    std = np.repeat(np.expand_dims(std,2),numwins,axis=2)
    pow_test_norm = np.subtract(pow_test,mean)/std

    feats_train = pow_train.reshape(len(perfs_train), -1)
    # train classifier first
    C = 0.0007
    alpha = 1/(2*C)
    ridge_classifier = Ridge(alpha=alpha)
    if perfnorm:
        ridge_classifier.fit(feats_train, perfs_train)
    else:
        ridge_classifier.fit(feats_train, logitPerf(perfs_train))
    
    preds = []
    # test classifier on loop
    for timewin in range(numwins):
        feats_test = pow_test_norm[:,:,timewin].reshape(1, -1)

        # Predict performance of left-out list
        preds_ridge = ridge_classifier.predict(feats_test)
        preds.extend(preds_ridge)
        
        #preds = np.array([ridge_classifier.predict(pow_test_norm[:,:,timewin].reshape(1, -1)) for timewin in range(numwins)]).flatten()

    return perfs_test, preds, times, times_words


def classifyPredictSlidingWindow_peristim_lolo(result,result_peristim,result_avg,list_test,freqs='all',perfnorm=False,avglistpow=True):
    # For each list train classifier on all other lists over all sessions, test on left-out list
    from sklearn.linear_model import Ridge
    
    # Concatenate all the lists and performances over all sessions together
    nsessions = len(result['perfs_list'])
    listnums = np.concatenate(result['lists'])
    if avglistpow:
        listPows = np.concatenate([res for res in result_avg['avgListPows_raw']])
    else:
        listPows = np.concatenate([result['listPows_raw'][s] for s in range(nsessions)])
    periStimPows = np.concatenate([result_peristim['itemPows_raw'][s] for s in range(nsessions)]) # concatentating the lists assumes equal number of time points per list/session
    perfs_all = np.concatenate(result['perfs_list'])
    
    # Remove electrodes with inf values
    goodelec1 = np.arange(listPows.shape[1])[[(e not in np.where(np.isinf(listPows))[1]) and
                                              (e not in np.where(np.isnan(listPows))[1]) for e in np.arange(listPows.shape[1])]]
    goodelec2 = np.arange(periStimPows.shape[1])[[(e not in np.where(np.isinf(periStimPows))[1]) and
                                                  (e not in np.where(np.isnan(periStimPows))[1]) for e in np.arange(periStimPows.shape[1])]]
    goodelec = np.array(list(set(goodelec1).intersection(set(goodelec2))))
    listPows = listPows[:,goodelec,:]
    periStimPows = periStimPows[:,goodelec,:,:]
    
    # Make sure frequencies align
    if freqs == 'trunc':
        listPows = listPows[:,:,-7:]
        periStimPows = periStimPows[:,:,-7:,:]
    
    # Reshape periStim array to be list x item x elec x freq x time
    periStimPows_reshaped = periStimPows.reshape(int(periStimPows.shape[0]/12),12,periStimPows.shape[1],periStimPows.shape[2],periStimPows.shape[3])
    
    # Arrange lists
    lists = np.arange(len(listPows))
    lists_train = lists[[l != list_test for l in lists]]
    
    # Identify which session the test list came from and normalize training lists by session
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    listPows_znorm, mean, std, set_omit = zscore_omit_list(listPows,listNumsBySess,omit=list_test) # this already has the test list omitted
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    perfs_all_norm, perfmean, perfstd, _ = norm_omit_list(logitPerf(perfs_all),listNumsBySess,omit=list_test)
    
    perfs_test = perfs_all[list_test]
    
    # Arrange training and test performances
    if perfnorm:
        perfs_train = perfs_all_norm # This is already logit transformed
        perfs_test = logitPerf([perfs_test])-perfmean # adjust by the mean of the set from which it came
    else:
        perfs_train = perfs_all[lists_train] # this will be logit transformed later
        perfs_test = logitPerf([perfs_test])
    
    listnum_test = listnums[list_test]
    recalls_test = result_peristim['recalls'][set_omit][result_peristim['recalls'][set_omit]['list']==listnum_test]
    recalled_test = list(recalls_test['recalled'])
    times = result_peristim['times_final']
    
    # apply norm to sliding windowed power
    pow_train = listPows_znorm
    pow_test = periStimPows_reshaped[list_test]
    numwins = periStimPows_reshaped.shape[-1]
    # expand mean and std over time and items
    mean = np.repeat(np.expand_dims(mean,2),numwins,axis=2) # over time
    mean = np.repeat(np.expand_dims(mean,0),12,axis=0) # over items
    std = np.repeat(np.expand_dims(std,2),numwins,axis=2)
    std = np.repeat(np.expand_dims(std,0),12,axis=0)
    pow_test_norm = np.subtract(pow_test,mean)/std

    feats_train = pow_train.reshape(len(perfs_train), -1)
    # train classifier first
    C = 0.0007
    alpha = 1/(2*C)
    ridge_classifier = Ridge(alpha=alpha)
    if perfnorm:
        ridge_classifier.fit(feats_train, perfs_train)
    else:
        ridge_classifier.fit(feats_train, logitPerf(perfs_train))
    
    preds = [[] for i in range(12)]
    # test classifier on loop for each item
    for it in range(12):
        for timewin in range(numwins):
            feats_test = pow_test_norm[it,:,:,timewin].reshape(1, -1)

            # Predict performance of left-out list
            preds_ridge = ridge_classifier.predict(feats_test)
            preds[it].extend(preds_ridge)

            #preds[it] = np.array([ridge_classifier.predict(pow_test_norm[it,:,:,timewin].reshape(1, -1)) for timewin in range(numwins)]).flatten()

    # then average over items in the list
    preds_mean = np.nanmean(np.array(preds),0)
    
    return perfs_test, preds_mean, times


def classifyPredictSlidingWindowByTime_peristim_lolo(result,list_test,freqs='all',perfnorm=False):
    # For each list train classifier on all other lists over all sessions, test on left-out list
    from sklearn.linear_model import Ridge
    
    # Concatenate all the lists and performances over all sessions together
    nsessions = len(result['perfs_list'])
    listnums = np.concatenate(result['lists'])
    #listPows = np.concatenate([result['listPows_raw'][s] for s in range(nsessions)])
    periStimPows = np.concatenate([result['itemPows_raw'][s] for s in range(nsessions)]) # concatentating the lists assumes equal number of time points per list/session
    
    perfs_all = np.concatenate(result['perfs_list'])
    
    # Remove electrodes with inf values
    #goodelec1 = np.arange(listPows.shape[1])[[(e not in np.where(np.isinf(listPows))[1]) and
    #                                          (e not in np.where(np.isnan(listPows))[1]) for e in np.arange(listPows.shape[1])]]
    goodelec = np.arange(periStimPows.shape[1])[[(e not in np.where(np.isinf(periStimPows))[1]) and
                                                  (e not in np.where(np.isnan(periStimPows))[1]) for e in np.arange(periStimPows.shape[1])]]
    #goodelec = np.array(list(set(goodelec1).intersection(set(goodelec2))))
    #listPows = listPows[:,goodelec,:]
    periStimPows = periStimPows[:,goodelec,:,:]
    
    # Make sure frequencies align
    if freqs == 'trunc':
        listPows = listPows[:,:,-7:]
        periStimPows = periStimPows[:,:,-7:,:]
    
    # Reshape periStim array to be list x item x elec x freq x time
    periStimPows_reshaped = periStimPows.reshape(int(periStimPows.shape[0]/12),12,periStimPows.shape[1],periStimPows.shape[2],periStimPows.shape[3])
    avgPows = np.nanmean(periStimPows_reshaped,axis=(1,4)) # average over item and time
    
    # Arrange lists
    lists = np.arange(len(periStimPows_reshaped))
    lists_train = lists[[l != list_test for l in lists]]
    
    timewins = result['times_final'][0:31]       

    perfs_test = perfs_all[list_test]
    
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    avgPows, mean, std, set_omit = zscore_omit_list(avgPows,listNumsBySess,omit=list_test) # this already has the test list omitted
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    perfs_all_norm, perfmean, perfstd, set_omit = norm_omit_list(logitPerf(perfs_all),listNumsBySess,omit=list_test)

    # Extend the power mean through time and item
    mean_time = np.repeat(np.expand_dims(np.repeat(np.expand_dims(mean,2),numwins,axis=2),0),12,axis=0)
    std_time = np.repeat(np.expand_dims(np.repeat(np.expand_dims(std,2),numwins,axis=2),0),12,axis=0)

    # Normalize the periStim by the mean
    
    # Arrange training and test performances
    if perfnorm:
        perfs_train = perfs_all_norm # This is already logit transformed
        perfs_test = logitPerf([perfs_test])-perfmean # adjust by the mean of the set from which it came
    else:
        perfs_train = perfs_all[lists_train] # this will be logit transformed later
        perfs_test = logitPerf([perfs_test])
    
    listnum_test = listnums[list_test]
    recalls_test = result['recalls'][set_omit][result['recalls'][set_omit]['list']==listnum_test]
    recalled_test = list(recalls_test['recalled'])

    # Start sub-loop for specific time cell
    preds = np.zeros((12,len(timewins)))
    preds[:] = np.nan

    for it in range(12):
        for t,timewin in enumerate(timewins):
            # subset the data into a specific item number and time window relative to that item number
            subListPows = periStimPows_reshaped[:,it,:,:,t] # list x elec x freq
            # Identify which session the test list came from and normalize training lists by session
            listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
            subListPows_znorm, mean, std, set_omit = zscore_omit_list(subListPows,listNumsBySess,omit=list_test) # this already has the test list omitted

            # apply norm to sliding windowed power
            pow_train = subListPows_znorm
            pow_test = periStimPows_reshaped[list_test,it,:,:,t]
            pow_test_norm = np.subtract(pow_test,mean_time)/std_time

            feats_train = pow_train.reshape(len(perfs_train), -1)
            feats_test = pow_test_norm.reshape(1, -1)

            # train classifier
            C = 0.0007
            alpha = 1/(2*C)
            ridge_classifier = Ridge(alpha=alpha)
            if perfnorm:
                ridge_classifier.fit(feats_train, perfs_train)
            else:
                ridge_classifier.fit(feats_train, logitPerf(perfs_train))
            
            # Predict performance of left-out list
            preds_ridge = ridge_classifier.predict(feats_test)
            preds[it,t] = preds_ridge


    # then average over items in the list, to get time course of classifier predictions
    preds_mean = np.nanmean(np.array(preds),0)
    
    return perfs_test, preds_mean


def getItemByListCorr(result):
    # Correlate the performance/power correlation vectors between item- and list-level power
    import numpy as np
    # Get item and list powers
    nsessions = len(result['perfs_list'])
    itemPows = np.concatenate([result['itemPows'][s] for s in range(nsessions)])
    listPows = np.concatenate([result['listPows'][s] for s in range(nsessions)])
    
    # Remove electrodes with inf values from either list or item
    goodelec_item = np.arange(itemPows.shape[1])[[(e not in np.where(np.isinf(itemPows))[1]) and
                                                   (e not in np.where(np.isnan(itemPows))[1]) for e in np.arange(itemPows.shape[1])]]
    goodelec_list = np.arange(listPows.shape[1])[[(e not in np.where(np.isinf(listPows))[1]) and
                                                   (e not in np.where(np.isnan(listPows))[1]) for e in np.arange(listPows.shape[1])]]
    goodelec = np.array(list(set(goodelec_item).intersection(set(goodelec_list))))
        
    # Remove electrodes with inf or nan values
    itemPows = itemPows[:,goodelec,:]
    listPows = listPows[:,goodelec,:]

    # Get performances
    perfs_all = np.concatenate(result['perfs_list'])
    perfs_all = logitPerf(perfs_all) # logit transform the performances
    recalls_all = np.concatenate([result['recalls'][s]['recalled'] for s in range(nsessions)])
    
    # reshape the power arrays
    itemPows_reshape = itemPows.reshape(len(recalls_all), -1)
    listPows_reshape = listPows.reshape(len(perfs_all), -1)
    
    # Correlate the power features with recalls
    corr_item = np.array([np.corrcoef(itemPows_reshape[:,i],recalls_all)[0,1] for i in range(itemPows_reshape.shape[1])])
    corr_item = corr_item.reshape(itemPows.shape[1:])
    
    # Correlate the power features with list performances
    corr_list = np.array([np.corrcoef(listPows_reshape[:,i],perfs_all)[0,1] for i in range(listPows_reshape.shape[1])])
    corr_list = corr_list.reshape(listPows.shape[1:])

    # Correlate the correlations
    itemListCorr = np.corrcoef(corr_item.flatten(),corr_list.flatten())[0,1]
    
    # And by frequency separately
    itemListCorr_byfreq = np.array([np.corrcoef(corr_item[:,f],corr_list[:,f])[0,1] for f in range(corr_item.shape[1])])
    
    return itemListCorr, itemListCorr_byfreq


def getCorrsByROI(result,scale='list',square=False,roinames=['ifg','mfg','sfg','tc','hc','phg','ipc','spc','oc'],hem='both',permTest=False):
    # Correlate power with performance and return aggregate ROI results
    # return a list of correlation for each contact and the contact's roi
    import numpy as np
    from localization import getMacroROIs_bilateral, getCoordVars, getHem
    
    # define rois of interest
    #roinames = ['ifg','mfg','sfg','tc','hc','phg','ipc','spc','oc']
    # define frequencies as string
    freqstr = [str(f) for f in result['myfreqs']]

    # get info on lists
    nsessions = len(result['perfs_list'])
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    
    # Get power
    nsessions = len(result['perfs_list'])
    if scale == 'item':
        pows = np.concatenate([result['itemPows'][s] for s in range(nsessions)])
        perfs = np.concatenate([result['recalls'][s]['recalled'] for s in range(nsessions)])
    elif scale == 'list':
        pows = np.concatenate([result['listPows'][s] for s in range(nsessions)])
        perfs = np.concatenate(result['perfs_list'])
        perfs = norm_list(logitPerf(perfs),listNumsBySess) # first logit transform and then normalize (de-mean) performances

    # Remove electrodes with inf values
    goodelec = np.arange(pows.shape[1])[[e not in np.where(np.isinf(pows))[1] and
                                         e not in np.where(np.isnan(pows))[1] for e in np.arange(pows.shape[1])]]

    pows = pows[:,goodelec,:]

    # reshape the power array
    pows_reshape = pows.reshape(len(perfs), -1)

    # Correlate the power features with list performances
    corr = np.array([np.corrcoef(pows_reshape[:,i],perfs)[0,1] for i in range(pows_reshape.shape[1])])
    corr = corr.reshape(pows.shape[1:]) # elec x freq

    if square:
        corr *= corr # Square it before averaging over contacts
    
    # Get the regions
    pairs = result['pairs'].iloc()[goodelec]
    regions = result['pairs_regions'][0][goodelec]
    hems = getHem(pairs)
    
    # Average the correlations across the ROIs for the selected hemisphere
    if hem == 'left':
        pairs = pairs[[h == 'l' for h in hems]]
        regions = regions[[h == 'l' for h in hems]]
        corr = corr[[h == 'l' for h in hems]]
    elif hem == 'right':
        pairs = pairs[[h == 'r' for h in hems]]
        regions = regions[[h == 'r' for h in hems]]
        corr = corr[[h == 'r' for h in hems]]
    
    # Average the correlations over ROIs
    rois = np.array(getMacroROIs_bilateral(regions))
    corrByROI = np.array([np.mean(corr[rois==roi],axis=0) for roi in roinames])
    
    nperm = 1000
    corrsByROI_shuff = []
    corrByROI_z = corrByROI
    if permTest:
        corrsByROI_shuff.append(corrByROI)
        # Calculate the z-score of the correlation based on permutation-based null distribution
        for n in range(nperm):
            # Shuffle the region labels
            rois_shuff = np.random.permutation(rois)
            # Then recompute the average
            corrByROI_shuff = np.array([np.mean(corr[rois_shuff==roi],axis=0) for roi in roinames])
            corrsByROI_shuff.append(corrByROI_shuff)
        corrsByROI_shuff = np.array(corrsByROI_shuff)
        corrByROI_z = zscore(corrsByROI_shuff,0)[0]
    
    # Also get the correlations for each electrode, with coordinate
    # Get the right coordinate variable
    corrByCoord = pd.DataFrame()
    coordvars = getCoordVars(pairs.iloc()[0],loc='avg')
    corrByCoord[['x','y','z']] = pairs[[coordvars[0],coordvars[1],coordvars[2]]]
    corrByCoord[freqstr] = corr # record correlation for each frequency
    
    return roinames, corrByROI, corrByROI_z, corrsByROI_shuff, corrByCoord, np.unique(rois)


def getCorrsByROI_avgpow(result,square=False,roinames=['ifg','mfg','sfg','tc','hc','phg','ipc','spc','oc'],hem='both',permTest=False):
    # Correlate power with performance and return aggregate ROI results; return a list of correlation for each contact and the contact's roi; Takes the average list power as input
    import numpy as np
    from scipy.stats import zscore
    from localization import getMacroROIs_bilateral, getCoordVars, getHem
    
    # define frequencies as string
    freqstr = [str(f) for f in result['myfreqs']]

    # get info on lists
    nsessions = len(result['perfs_list'])
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    
    # Get power
    nsessions = len(result['perfs_list'])
    pows = np.concatenate([zscore(result['avgListPows_raw'][s]) for s in range(nsessions)])
    perfs = np.concatenate(result['perfs_list'])
    perfs = norm_list(logitPerf(perfs),listNumsBySess) # first logit transform and then normalize (de-mean) performances

    # Remove electrodes with inf values
    goodelec = np.arange(pows.shape[1])[[e not in np.where(np.isinf(pows))[1] and
                                         e not in np.where(np.isnan(pows))[1] for e in np.arange(pows.shape[1])]]

    pows = pows[:,goodelec,:]

    # reshape the power array
    pows_reshape = pows.reshape(len(perfs), -1)

    # Correlate the power features with list performances
    corr = np.array([np.corrcoef(pows_reshape[:,i],perfs)[0,1] for i in range(pows_reshape.shape[1])])
    corr = corr.reshape(pows.shape[1:]) # elec x freq

    if square:
        corr *= corr # Square it before averaging over contacts
    
    # Get the regions
    pairs = result['pairs'].iloc()[goodelec]
    regions = result['pairs_regions'][0][goodelec]
    hems = getHem(pairs)
    
    # Average the correlations across the ROIs for the selected hemisphere
    if hem == 'left':
        pairs = pairs[[h == 'l' for h in hems]]
        regions = regions[[h == 'l' for h in hems]]
        corr = corr[[h == 'l' for h in hems]]
    elif hem == 'right':
        pairs = pairs[[h == 'r' for h in hems]]
        regions = regions[[h == 'r' for h in hems]]
        corr = corr[[h == 'r' for h in hems]]
    
    # Average the correlations over ROIs
    rois = np.array(getMacroROIs_bilateral(regions))
    corrByROI = np.array([np.mean(corr[rois==roi],axis=0) for roi in roinames])
    
    nperm = 1000
    corrsByROI_shuff = []
    corrByROI_z = corrByROI
    if permTest:
        corrsByROI_shuff.append(corrByROI)
        # Calculate the z-score of the correlation based on permutation-based null distribution
        for n in range(nperm):
            # Shuffle the region labels
            rois_shuff = np.random.permutation(rois)
            # Then recompute the average
            corrByROI_shuff = np.array([np.mean(corr[rois_shuff==roi],axis=0) for roi in roinames])
            corrsByROI_shuff.append(corrByROI_shuff)
        corrsByROI_shuff = np.array(corrsByROI_shuff)
        corrByROI_z = zscore(corrsByROI_shuff,0)[0]
    
    # Also get the correlations for each electrode, with coordinate
    # Get the right coordinate variable
    corrByCoord = pd.DataFrame()
    coordvars = getCoordVars(pairs.iloc()[0],loc='avg')
    corrByCoord[['x','y','z']] = pairs[[coordvars[0],coordvars[1],coordvars[2]]]
    corrByCoord[freqstr] = corr # record correlation for each frequency
    
    return roinames, corrByROI, corrByROI_z, corrsByROI_shuff, corrByCoord, np.unique(rois)


def getCorrsByROITimeFreq(subj,task,winsize=10000,roinames=['ifg','mfg','sfg','tc','hc','phg','ipc','spc','oc'],resdir='/home1/daniel.rubinstein/Projects/listSME/singlesubj/8freq_wholeList_-1_31/'):
    # Correlate power across regions by time and frequency
    # Takes as input result_wholeList, which is an average over overlapping time windows of a given size
    import numpy as np
    from scipy.stats import zscore
    from localization import getMacroROIs_bilateral, getCoordVars, getHem

    # Load data
    timestr = str(winsize)+'mswindows'
    result_wholeList = np.load(resdir+subj+'_'+task+'_wholeListpow_'+timestr+'.npy',allow_pickle=True).item()

    # Arrange frequencies and times
    myfreqs = result_wholeList['myfreqs']
    
    if winsize==1000:
        timewins = result_wholeList['times_final'][0:311]
    elif winsize==3000:
        timewins = result_wholeList['times_final'][0:97]
    elif winsize==6000:
        timewins = result_wholeList['times_final'][0:44]
    elif winsize==10000:
        timewins = result_wholeList['times_final'][0:23]
    elif winsize==15000:
        timewins = result_wholeList['times_final'][0:12]

    # z-score the power and concatenate
    wholeListPows_znorm = np.concatenate([zscore(pows,axis=0) for pows in result_wholeList['wholeListPows_raw']])

    # Remove electrodes with inf and nan values
    goodelec = np.arange(wholeListPows_znorm.shape[1])[[(e not in np.where(np.isinf(wholeListPows_znorm))[1]) and
                                             (e not in np.where(np.isnan(wholeListPows_znorm))[1]) for e in np.arange(wholeListPows_znorm.shape[1])]]

    wholeListPows_znorm = wholeListPows_znorm[:,goodelec,:,:]
    perfs_all = np.concatenate([perfs-np.mean(perfs) for perfs in result_wholeList['perfs_list']])

    # Start sub-loop for specific time-frequency cell
    corrs = np.zeros((len(goodelec),len(timewins),len(myfreqs)))
    corrs[:] = np.nan

    for t,f in [(x,y) for x,_ in enumerate(timewins) for y,_ in enumerate(myfreqs)]:
        subListPows_znorm = wholeListPows_znorm[:,:,f,t]

        # Calculate correlations between power and performance for all electrodes at once
        corr = np.array([np.corrcoef(subListPows_znorm[:,i],perfs_all)[0,1] for i in range(subListPows_znorm.shape[1])])
        corrs[:,t,f] = corr

    # Average the correlations over ROIs
    regions = result_wholeList['pairs_regions'][0][goodelec]
    rois = np.array(getMacroROIs_bilateral(regions))
    corrsByROI = np.array([np.mean(corrs[rois==roi],axis=0) for roi in roinames])


    return roinames, corrs, corrsByROI


def classifyListPredictItems_lolo(result,list_test,permute=False,perfnorm=False):
    # For each list train classifier on all other lists over all sessions, test on left-out list
    from sklearn.linear_model import LogisticRegression, Ridge
    from scipy.stats import zscore
    
    # Concatenate all the lists and performances over all sessions together
    nsessions = len(result['perfs_list'])
    listPows = np.concatenate([result['listPows_raw'][s] for s in range(nsessions)])
    itemPows = np.concatenate([result['itemPows_raw'][s] for s in range(nsessions)])
    
    # Remove electrodes with inf and nan values
    goodelec = np.arange(listPows.shape[1])[[(e not in np.where(np.isinf(listPows))[1]) and
                                             (e not in np.where(np.isnan(listPows))[1]) for e in np.arange(listPows.shape[1])]]

    listPows = listPows[:,goodelec,:]
    itemPows = itemPows[:,goodelec,:]
    
    recalls_all = np.concatenate([result['recalls'][s]['recalled'] for s in range(nsessions)])
    perfs_all = np.concatenate(result['perfs_list'])
    #perfs_all_norm = np.concatenate([perfs-np.mean(perfs) for perfs in result['perfs_list']])
    
    lists = np.arange(len(listPows))
    lists_train = lists[[l != list_test for l in lists]]
    
    # Identify which session the test list came from and normalize training lists by session
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    listPows_znorm, mean, std, set_omit = zscore_omit_list(listPows,listNumsBySess,omit=list_test) # this already has the test list omitted
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    perfs_all_norm, perfmean, perfstd, _ = norm_omit_list(logitPerf(perfs_all),listNumsBySess,omit=list_test)
    
    # Identify which session the test list came from and normalize items by session
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    itemPows_reshaped = itemPows.reshape(int(itemPows.shape[0]/12),12,itemPows.shape[1],itemPows.shape[2]) # reshape by list
    itemPows_znorm, mean_item, mean_std, set_omit = zscore_omit_item(itemPows_reshaped,listNumsBySess,omit=list_test) # output has test list omitted
    
    recalls_test = recalls_all[12*list_test:12*(list_test+1)]
    #perfs_test = perfs_all[list_test]
    
    if perfnorm:
        perfs_train = perfs_all_norm # this is already logit-transformed
        #perfs_test = logitPerf([perfs_test])-perfmean # adjust by the mean from the set which it came from
    else:
        perfs_train = perfs_all[lists_train] # this will be logit-transformed later
        #perfs_test = logitPerf([perfs_test])
    if permute:
        np.random.shuffle(perfs_train)

    pow_train = listPows_znorm
    #pow_test = listPows[list_test]
    #pow_test_norm = np.subtract(pow_test,mean)/std
    pow_test = itemPows_reshaped[list_test]
    pow_test_norm = np.subtract(pow_test,mean_item)/mean_std
    
    feats_train = pow_train.reshape(len(perfs_train), -1)
    #feats_test = pow_test_norm.reshape(1, -1)
    feats_test = pow_test_norm.reshape(12, -1)
        
    # Ridge Regression Classifier
    C = 0.0007
    alpha = 1/(2*C)
    ridge_classifier = Ridge(alpha=alpha)
    if perfnorm:
        ridge_classifier.fit(feats_train, perfs_train)
    else:
        ridge_classifier.fit(feats_train, logitPerf(perfs_train))

    # Predict recalls of items from left-out list
    preds = ridge_classifier.predict(feats_test)
    
    # Calculate ROC and AUC
    #preds_probs = ridge_classifier.predict_proba(feats_test)[:,1]
    #auc = roc_auc_score(recalls_test, preds_probs)
    #roc = roc_curve(recalls_test, preds_probs)

    return recalls_test, preds


def classifyItemsPredictList_lolo(result,list_test,permute=False,balanced_weights=True,perfnorm=True):
    # For each list train classifier on all other lists over all sessions, test on left-out list
    from sklearn.linear_model import LogisticRegression
    from sklearn.metrics import roc_auc_score, roc_curve
    from scipy.stats import zscore
    
    # Concatenate all the lists and performances over all sessions together
    nsessions = len(result['perfs_list'])
    itemPows = np.concatenate([result['itemPows_raw'][s] for s in range(nsessions)])
    listPows = np.concatenate([result['listPows_raw'][s] for s in range(nsessions)])
    
    # Remove electrodes with inf values
    goodelec = np.arange(itemPows.shape[1])[[(e not in np.where(np.isinf(itemPows))[1]) and
                                             (e not in np.where(np.isnan(itemPows))[1]) for e in np.arange(itemPows.shape[1])]]

    itemPows = itemPows[:,goodelec,:]
    listPows = listPows[:,goodelec,:]
    
    recalls_all = np.concatenate([result['recalls'][s]['recalled'] for s in range(nsessions)])
    perfs_all = np.concatenate(result['perfs_list'])
    lists = np.arange(len(perfs_all))
    lists_train = lists[[l != list_test for l in lists]]
    items = np.arange(len(itemPows))
    items_train = items[[i not in np.arange(12*list_test,12*(list_test+1)) for i in items]]
    
    # Identify which session the test list came from and normalize items by session
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    itemPows_reshaped = itemPows.reshape(int(itemPows.shape[0]/12),12,itemPows.shape[1],itemPows.shape[2]) # reshape by list
    itemPows_znorm, mean, std, set_omit = zscore_omit_item(itemPows_reshaped,listNumsBySess,omit=list_test) # output has test list omitted
    
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    listPows_znorm, mean_list, std_list, set_omit = zscore_omit_list(listPows,listNumsBySess,omit=list_test) # this already has the test list omitted
    listNumsBySess = [len(result['perfs_list'][s]) for s in range(nsessions)]
    perfs_all_norm, perfmean, perfstd, _ = norm_omit_list(logitPerf(perfs_all),listNumsBySess,omit=list_test)
    
    recalls_train = recalls_all[items_train]
    if permute:
        np.random.shuffle(recalls_train)
        
    #recalls_test = recalls_all[12*list_test:12*(list_test+1)]
    perfs_test = perfs_all[list_test]
    if perfnorm:
        perfs_test = logitPerf([perfs_test])-perfmean # adjust by the mean from the set which it came from
    else:
        perfs_test = logitPerf([perfs_test])
    
    pow_train = itemPows_znorm
    #pow_test = itemPows_reshaped[list_test]
    pow_test = listPows[list_test]
    pow_test_norm = np.subtract(pow_test,mean_list)/std_list
    
    feats_train = pow_train.reshape(len(recalls_train), -1)
    #feats_test = pow_test_norm.reshape(12, -1)
    feats_test = pow_test_norm.reshape(1, -1)
    
    # Logistic Regression Classifier
    C = 0.0007
    if balanced_weights:
        lr_classifier = LogisticRegression(C=C, penalty='l2', solver='liblinear', multi_class='ovr', n_jobs=1,
                                      class_weight='balanced')
    else:
        lr_classifier = LogisticRegression(C=C, penalty='l2', solver='liblinear', multi_class='ovr', n_jobs=1)
    lr_classifier.fit(feats_train, recalls_train)
    
    # Predict list (good or bad) and probability
    preds = lr_classifier.predict(feats_test)
    preds_probs = lr_classifier.predict_proba(feats_test)[:,1]
    
    return perfs_test, preds, preds_probs


def getRecTimes(subj,task):
    from eegQC import isBadSession
    rectimes = []
    #Find out the sessions
    df = get_data_index("r1")
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])
    for s, sess in enumerate(sessions):
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s]) #reader for loading CML data
        if isBadSession(subj,task,sess):
            continue
        evs = reader.load('task_events')
        evs = evs[evs['eegoffset'] > -1]
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        list_evs = evs[(evs['type']=='TRIAL') & (evs['list']>0)]
        words_first = word_evs[word_evs['serialpos']==1]

        # Only use lists of length 12 and also has a recall period
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete_sess = word_evs['list'].unique()[[n==12 for n in nwords]]
        lists_withrec = evs[evs['type']=='REC_START']['list'].unique()
        lists_withend = evs[evs['type']=='REC_END']['list'].unique()
        lists_complete_sess = list(set(lists_withrec).intersection(set(lists_complete_sess)))
        if subj == 'R1342M':
            lists_complete_sess = list(set(lists_withend).intersection(set(lists_complete_sess)))
        if ((subj == 'R1395M') & (task == 'catFR1') & (sess == 1)):
            lists_complete_sess.remove(10)
        word_evs = word_evs[[l in lists_complete_sess for l in word_evs['list']]]
        list_evs = list_evs[[l in lists_complete_sess for l in list_evs['list']]]
        words_first = words_first[[l in lists_complete_sess for l in words_first['list']]]

        recstarts = evs[evs['type']=='REC_START']
        recstarts = recstarts[[l in lists_complete_sess for l in recstarts['list']]]
        recstarts_rel = np.subtract(list(recstarts['mstime']),list(words_first['mstime']))

        rectimes.extend(recstarts_rel)
    
    return rectimes


def getDistractorTimes(subj,task,disttype='start'):
    from eegQC import isBadSession
    disttimes = []
    #Find out the sessions
    df = get_data_index("r1")
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])
    for s, sess in enumerate(sessions):
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s]) #reader for loading CML data
        if isBadSession(subj,task,sess):
            continue
        evs = reader.load('task_events')
        evs = evs[evs['eegoffset'] > -1]
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        list_evs = evs[(evs['type']=='TRIAL') & (evs['list']>0)]
        words_first = word_evs[word_evs['serialpos']==1]

        # Only use lists of length 12 and also has a recall period
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete_sess = word_evs['list'].unique()[[n==12 for n in nwords]]
        lists_withrec = evs[evs['type']=='REC_START']['list'].unique()
        lists_withend = evs[evs['type']=='REC_END']['list'].unique()
        lists_complete_sess = list(set(lists_withrec).intersection(set(lists_complete_sess)))
        if subj == 'R1342M':
            lists_complete_sess = list(set(lists_withend).intersection(set(lists_complete_sess)))
        if ((subj == 'R1395M') & (task == 'catFR1') & (sess == 1)):
            lists_complete_sess.remove(10)
        word_evs = word_evs[[l in lists_complete_sess for l in word_evs['list']]]
        list_evs = list_evs[[l in lists_complete_sess for l in list_evs['list']]]
        words_first = words_first[[l in lists_complete_sess for l in words_first['list']]]

        if disttype=='start':
            diststarts = evs[evs['type']=='DISTRACT_START']
        else:
            diststarts = evs[evs['type']=='DISTRACT_END']
        diststarts = diststarts[[l in lists_complete_sess for l in diststarts['list']]]
        diststarts_rel = np.subtract(list(diststarts['mstime']),list(words_first['mstime']))

        disttimes.extend(diststarts_rel)
    
    return disttimes


def getRecEndTimes(subj,task):
    from eegQC import isBadSession
    rec_end_times = []
    #Find out the sessions
    df = get_data_index("r1")
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])
    for s, sess in enumerate(sessions):
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s]) #reader for loading CML data
        if isBadSession(subj,task,sess):
            continue
        evs = reader.load('task_events')
        evs = evs[evs['eegoffset'] > -1]
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        list_evs = evs[(evs['type']=='TRIAL') & (evs['list']>0)]
        words_first = word_evs[word_evs['serialpos']==1]

        # Only use lists of length 12 and also has a recall period
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete_sess = word_evs['list'].unique()[[n==12 for n in nwords]]
        lists_withrec = evs[evs['type']=='REC_START']['list'].unique()
        lists_withend = evs[evs['type']=='REC_END']['list'].unique()
        lists_complete_sess = list(set(lists_withrec).intersection(set(lists_complete_sess)))
        if subj == 'R1342M':
            lists_complete_sess = list(set(lists_withend).intersection(set(lists_complete_sess)))
        if ((subj == 'R1395M') & (task == 'catFR1') & (sess == 1)):
            lists_complete_sess.remove(10)
        word_evs = word_evs[[l in lists_complete_sess for l in word_evs['list']]]
        list_evs = list_evs[[l in lists_complete_sess for l in list_evs['list']]]
        words_first = words_first[[l in lists_complete_sess for l in words_first['list']]]
        
        countdowns = evs[(evs['type']=='COUNTDOWN_START') & (evs['list']>0)]
        countdowns = countdowns[[l in lists_complete_sess for l in countdowns['list']]]
        
        for ll in lists_complete_sess[1:]:
            # For each list following a recall period, get the closest countdown period within 30 seconds of the start of the encoding period
            enc_start = words_first[words_first['list']==ll]['mstime'].iloc()[0] # mstime of first word
            countdown_opts = countdowns[countdowns['list']<=ll]
            # take the most recent countdown occuring BEFORE the encoding start
            difftimes = np.subtract(list(countdown_opts['mstime']),enc_start)
            countdown_opts = countdown_opts.iloc()[difftimes<0]
            cdstart = countdown_opts.iloc()[-1]['mstime']
            # Get the first rec_end event with 20 seconds of the countdown start
            rec_ends_mstimes = np.array(evs[evs['type']=='REC_END']['mstime'])
            rec_ends_mstimes_opts = rec_ends_mstimes[(rec_ends_mstimes-cdstart<0) & (rec_ends_mstimes-cdstart>-20000)]
            if len(rec_ends_mstimes_opts)>0:
                rec_end_mstime = rec_ends_mstimes_opts[0]
                rec_end_time = enc_start-rec_end_mstime
                rec_end_times.append(rec_end_time)
            
    return rec_end_times


def getCountdownToEncTimes(subj,task):
    from eegQC import isBadSession
    cdtimes = []
    #Find out the sessions
    df = get_data_index("r1")
    sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
    monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
    locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])
    for s, sess in enumerate(sessions):
        reader = CMLReader(subj, task, sessions[s], montage=monts[s], localization=locs[s]) #reader for loading CML data
        if isBadSession(subj,task,sess):
            continue
        evs = reader.load('task_events')
        evs = evs[evs['eegoffset'] > -1]
        word_evs = evs[(evs['type']=='WORD') & (evs['list']>0)]
        list_evs = evs[(evs['type']=='TRIAL') & (evs['list']>0)]
        words_first = word_evs[word_evs['serialpos']==1]

        # Only use lists of length 12 and also has a recall period
        nwords = [len(word_evs[word_evs['list']==l]) for l in word_evs['list'].unique()]
        lists_complete_sess = word_evs['list'].unique()[[n==12 for n in nwords]]
        lists_withrec = evs[evs['type']=='REC_START']['list'].unique()
        lists_withend = evs[evs['type']=='REC_END']['list'].unique()
        lists_complete_sess = list(set(lists_withrec).intersection(set(lists_complete_sess)))
        if subj == 'R1342M':
            lists_complete_sess = list(set(lists_withend).intersection(set(lists_complete_sess)))
        if ((subj == 'R1395M') & (task == 'catFR1') & (sess == 1)):
            lists_complete_sess.remove(10)
        word_evs = word_evs[[l in lists_complete_sess for l in word_evs['list']]]
        list_evs = list_evs[[l in lists_complete_sess for l in list_evs['list']]]
        words_first = words_first[[l in lists_complete_sess for l in words_first['list']]]
        
        countdowns = evs[(evs['type']=='COUNTDOWN_START') & (evs['list']>0)]
        countdowns = countdowns[[l in lists_complete_sess for l in countdowns['list']]]
        
        for ll in lists_complete_sess:
            # For each list, get the closest countdown period within 30 seconds of the start of the encoding period
            enc_start = words_first[words_first['list']==ll]['mstime'].iloc()[0] # mstime of first word
            countdown_opts = countdowns[countdowns['list']<=ll]
            # take the most recent countdown occuring BEFORE the encoding start
            difftimes = np.subtract(list(countdown_opts['mstime']),enc_start)
            countdown_opts = countdown_opts.iloc()[difftimes<0]
            cdstart = countdown_opts.iloc()[-1]['mstime']
            cdtime = enc_start-cdstart
            if cdtime < 30000: # only append if time is less than 30 seconds
                cdtimes.append(cdtime)
    
    return cdtimes

