#Module to hold useful functions regarding localizations and regions, etc.
import numpy as np
import pandas as pd
from cmlreaders import CMLReader, get_data_index


def getMacroROIs_bilateral(regions):
    # Function to take in array of specific regions as defined by the get_elec_regions function below,
    # and return a macro region from among: [IFG, MFG, SFG, TC, HIP, PHC, IPC, SPC, OC] as in Kragel/Weidemann et al 2019
    
    regions = np.array(regions)
    roinames = np.array(['ifg','mfg','sfg','tc','hc','phg','ipc','spc','oc'])
    
    ifg = ['opifg opercular part of the inferior frontal gyrus','orifg orbital part of the inferior frontal gyrus','parsopercularis',
       'parsorbitalis','parstriangularis','trifg triangular part of the inferior frontal gyrus']
    mfg = ['caudal middle frontal cortex','caudalmiddlefrontal','dlpfc','mfg middle frontal gyrus','rostralmiddlefrontal']
    sfg = ['msfg superior frontal gyrus medial segment','sfg superior frontal gyrus','superiorfrontal']
    tc = ['bankssts','inferiortemporal','fug fusiform gyrus', 'fusiform', 'fusiform gyrus wm','itg inferior temporal gyrus',
       'middle temporal gyrus', 'middletemporal','mtg','mtg middle temporal gyrus','stg','stg superior temporal gyrus',
       'superiortemporal','temporalpole','tmp temporal pole', 'transversetemporal','ttg transverse temporal gyrus','tc']
    hc = ['ca1', 'ca2', 'ca3','dg','hippocampus','sub']
    phg = ['ba35', 'ba36','ec', 'ent entorhinal area', 'entorhinal','erc','parahippocampal','phc','phg parahippocampal gyrus','prc']
    ipc = ['ang angular gyrus','inferiorparietal','smg supramarginal gyrus','supramarginal gyrus','supramarginal']
    spc = ['spl superior parietal lobule','superiorparietal']
    oc = ['calc calcarine cortex','cun cuneus', 'cuneus','iog inferior occipital gyrus','lateraloccipital','lig lingual gyrus',
       'lingual', 'lorg lateral orbital gyrus','mog middle occipital gyrus','ocp occipital pole','ofug occipital fusiform gyrus',
       'pericalcarine','sog superior occipital gyrus']
    
    roilists = [ifg, mfg, sfg, tc, hc, phg, ipc, spc, oc]
    
    #roilists_left = [['left ' + r for r in roilist] for roilist in roilists]
    #roilists_right = [['right ' + r for r in roilist] for roilist in roilists]
    
    regions = [r.strip().strip('""') for r in regions]
    regions = [r.replace('left ','').replace('right ','') for r in regions]
    
    rois = []
    # for each region, check which roi it's in, if any
    for region in regions:
        roi = roinames[[region in roilist for roilist in roilists]]
        #roi_left = roinames[[region in roilist for roilist in roilists_left]]
        #roi_right = roinames[[region in roilist for roilist in roilists_right]]
        
        if len(roi) != 1:
            roi = ['none']
        
        rois.append(roi[0])
    

    return rois


def getMacroROIs_unilateral(regions,hemisphere='left'):
    # Function to take in array of specific regions as defined by the get_elec_regions function below,
    # and return a macro region from among: [IFG, MFG, SFG, TC, HIP, PHC, IPC, SPC, OC] as in Kragel/Weidemann et al 2019
    
    regions = np.array(regions)
    roinames = np.array(['ifg','mfg','sfg','tc','hc','phg','ipc','spc','oc'])
    
    ifg = ['opifg opercular part of the inferior frontal gyrus','orifg orbital part of the inferior frontal gyrus','parsopercularis',
       'parsorbitalis','parstriangularis','trifg triangular part of the inferior frontal gyrus']
    mfg = ['caudal middle frontal cortex','caudalmiddlefrontal','dlpfc','mfg middle frontal gyrus','rostralmiddlefrontal']
    sfg = ['msfg superior frontal gyrus medial segment','sfg superior frontal gyrus','superiorfrontal']
    tc = ['bankssts','inferiortemporal','fug fusiform gyrus', 'fusiform', 'fusiform gyrus wm','itg inferior temporal gyrus',
       'middle temporal gyrus', 'middletemporal','mtg','mtg middle temporal gyrus','stg','stg superior temporal gyrus',
       'superiortemporal','temporalpole','tmp temporal pole', 'transversetemporal','ttg transverse temporal gyrus','tc']
    hc = ['ca1', 'ca2', 'ca3','dg','hippocampus','sub']
    phg = ['ba35', 'ba36','ec', 'ent entorhinal area', 'entorhinal','erc','parahippocampal','phc','phg parahippocampal gyrus','prc']
    ipc = ['ang angular gyrus','inferiorparietal','smg supramarginal gyrus','supramarginal gyrus','supramarginal']
    spc = ['spl superior parietal lobule','superiorparietal']
    oc = ['calc calcarine cortex','cun cuneus', 'cuneus','iog inferior occipital gyrus','lateraloccipital','lig lingual gyrus',
       'lingual', 'lorg lateral orbital gyrus','mog middle occipital gyrus','ocp occipital pole','ofug occipital fusiform gyrus',
       'pericalcarine','sog superior occipital gyrus']
    
    roilists = [ifg, mfg, sfg, tc, hc, phg, ipc, spc, oc]
    rois = []
    # for each region, check which roi it's in, if any
    for region in regions:
        roi = roinames[[region in roilist for roilist in roilists]]
        if len(roi)!=1:
            roi = ['none']
        
        rois.append(roi[0])
    

    return rois


def getPriorityMTLrois(reader, elecs, checkbad=True):
    #Function to divide electrodes into either PHG or the Hippocampus
    roilist = list()
    
    #Initialize to 'none'
    for e in range(len(elecs)):
        roilist.append('none')
    
    mtllist_dk = ['parahippocampal','entorhinal','perirhinal']
    ltclist = ['superiortemporal','middletemporal','inferiortemporal','fusiform','bankssts','temporalpole','transversetemporal']
    
    itglist = ['inferiortemporal']
    mtglist = ['middletemporal']
    stglist = ['superiortemporal','bankssts']

    hipplist_stein = ['Left CA1','Left CA2','Left CA3','Left DG','Left Sub','Right CA1','Right CA2','Right CA3','Right DG','Right Sub']
    mtllist_stein = ['Left EC','Left PHC','Left PRC','Right EC','Right PHC','Right PRC']
    
    hipplist_das = ['CA1','CA2','CA3','DG','SUB']
    mtllist_das = ['BA35','BA36','ERC','PHC']
    
    hipplist_mtl_atlas = ['CA1','DG','SUB']
    mtllist_mtl_atlas = ['BA35','BA36','ERC','PHC']
    
    hipplist_all = ['CA1','CA2','CA3','DG','Sub','SUB']
    hipplist_all_lower = ['ca1','ca2','ca3','dg','sub']
    
    mtllist_all = ['BA35','BA36','EC','ERC','PHC','PRC']
    mtllist_all_lower = ['ba35','ba36','ec','erc','phc','prc']
    
    WMlist_stein = ['Left Fusiform Gyrus WM','Left MTL WM','Right MTL WM']
    WMlist_stein_lower = ['left fusiform gyrus wm','left mtl wm','right mtl wm']
    
    checks = np.ones(len(elecs)).astype(bool)
    if checkbad:
        checks = checkBadElectrodes(reader, elecs)
    
    #Update the elecs structure to include the new mtl atlas as field "mtl_atlas"
    try:
        localization = reader.load('localization')
        elecs = updateElecs(localization, elecs)
    except:
        pass
    
    for e in range(len(elecs)):
        elec = elecs.iloc()[e]

        #Initialize the different atlas regions
        steinreg = 'none'
        dasreg = 'none'
        mtlatlasreg = 'none'
        indreg = 'none'

        #First check if the electrode is not bad
        if checks[e]:

            #Check if different roi columns exists
            if 'stein.region' in elec.keys():
                steinreg = str(elec['stein.region'])
                steinreg = steinreg.strip('"').lower()

            if 'das.region' in elec.keys(): 
                dasreg = str(elec['das.region'])
                dasreg = dasreg.strip('"').lower()

            if 'mtl_atlas' in elec.keys():
                mtlatlasreg = str(elec['mtl_atlas'])
                mtlatlasreg = mtlatlasreg.strip('"').lower()

            if 'ind.region' in elec.keys():
                indreg = str(elec['ind.region'])
                indreg = indreg.strip('"').lower()

            # Do localization without prioritizing one method over the other:
            steinhipp = any([r in steinreg for r in hipplist_all_lower])
            steinphg = any([r in steinreg for r in mtllist_all_lower])
            dashipp = any([r in dasreg for r in hipplist_all_lower])
            dasphg = any([r in dasreg for r in mtllist_all_lower])
            mtlatlashipp = any([r in mtlatlasreg for r in hipplist_all_lower])
            mtlatlasphg = any([r in mtlatlasreg for r in mtllist_all_lower])
            indphg = any([r in indreg for r in mtllist_dk])
            indltc = any([r in indreg for r in ltclist])
            inditg = any([r in indreg for r in itglist])
            indmtg = any([r in indreg for r in mtglist])
            indstg = any([r in indreg for r in stglist])

            if steinhipp | dashipp | mtlatlashipp:
                roilist[e] = 'hipp'
            elif steinphg | dasphg | mtlatlasphg | indphg:
                roilist[e] = 'phg'
            elif inditg:
                roilist[e] = 'itg'
            elif indmtg:
                roilist[e] = 'mtg'
            elif indstg:
                roilist[e] = 'stg'
            elif indltc:
                roilist[e] = 'ltc'


    return np.array(roilist)


def checkBadElec(reader, elec):
    # Function to check if a single elec series is bad
    import pandas as pd
    from os import mkdir
    from cmlreaders import CMLReader
    
    elec_result = True
 
    try:
        elecs_cat = reader.load('electrode_categories')
        
    except:
        #print('No electrode categories file found!')
        return elec_result
    

    elec_labels = str(elec['label']).split('-')
    #try to account for dashes in name
    if (len(elec_labels) > 2) & (len(elec_labels) % 2 == 0):
        n2 = int(len(elec_labels)/2)
        elec_labels = ['-'.join(elec_labels[0:n2]), '-'.join(elec_labels[n2:])]
    for label in elec_labels:
        soz = label in elecs_cat['soz']
        interictal = label in elecs_cat['interictal']
        bad_channel = label in elecs_cat['bad_channel']
        brain_lesion = label in elecs_cat['brain_lesion']
        nanchan = label in ['nan']
        mancheck = manualCheck(reader.subject,label)
            
            #if soz or interictal or bad_channel or brain_lesion:
        if bad_channel or brain_lesion or nanchan or mancheck:
            elec_result = False
                  
    return elec_result


def checkBadElectrodes(reader, elecs):
    #Function to remove the electrodes from seizure onset zone and irritative zone (exhibiting spikes)
    #takes as arguments: reader and electrodes dataframe

    import pandas as pd
    from os import mkdir
    from cmlreaders import CMLReader
    
    elecs_result = np.ones(len(elecs)).astype(bool)
    
    if not isinstance(elecs, pd.DataFrame):
        print('Third argument must be a DataFrame!')
        return elecs_result
    
    #df = get_data_index("r1")
    #sessions = list(df[(df['subject']==sub) & (df['experiment']==task)]['session'])
    #mont = int(df[(df['subject']==sub) & (df['experiment']==task)].iloc()[0]['montage'])      
    #loc = int(df[(df['subject']==sub) & (df['experiment']==task)].iloc()[0]['localization'])
    #reader = CMLReader(sub, task, sessions[0], montage=mont, localization=loc)
    
    try:
        elecs_cat = reader.load('electrode_categories')
        
    except:
        #print('No electrode categories file found!')
        return elecs_result
    
    
    for e in range(len(elecs)):
        elec = elecs.iloc()[e]
        elec_labels = str(elec['label']).split('-')
        #try to account for dashes in name
        if (len(elec_labels) > 2) & (len(elec_labels) % 2 == 0):
            n2 = int(len(elec_labels)/2)
            elec_labels = ['-'.join(elec_labels[0:n2]), '-'.join(elec_labels[n2:])]
        for label in elec_labels:
            soz = label in elecs_cat['soz']
            interictal = label in elecs_cat['interictal']
            bad_channel = label in elecs_cat['bad_channel']
            brain_lesion = label in elecs_cat['brain_lesion']
            nanchan = label in ['nan']
            mancheck = manualCheck(reader.subject,label)
            
            #if soz or interictal or bad_channel or brain_lesion:
            if bad_channel or brain_lesion or nanchan or mancheck:
                elecs_result[e] = False
                
        
    #Return an array of length len(elecs) of either 'True' or 'False'
    return elecs_result


def manualCheck(subj,label):
    bads = []
    path = '/home1/daniel.rubinstein/Projects/dataQC/badChannels/'
    file = str(path+subj+'.txt')
    try:
        bads = [line.rstrip('\n') for line in open(file)]
    except:
        pass
        #print('No manually labeled bad channels for',subj)
        
    return str(label) in bads


def updateElec(localization, elec):
    #Stolen from Ethan Solomon
    #For recent (R1380+) subjects who have MTL elecs in localization file but not pairs file

    #First check if bipolar or monopolar
    if 'contact_1' in elec.keys(): #this is a pairs object
        elec_new = elec.copy()
        if 'atlases.mtl' in localization.keys():
            #Add the mtl atlas label to the original elecs structure
            for atlas_idx, pair in enumerate(localization['atlases.mtl']['pairs'].index):
                pairlabel = pair[0]+'-'+pair[1]
                try:
                    #Find where in the mtl atlas the electrode is located
                    if elec_new['label'] == pairlabel:
                        elec_new['mtl_atlas'] = localization['atlases.mtl']['pairs'][atlas_idx]
                except:
                    continue
                
            return elec_new
        else:
            print('No MTL information found!')
            return elec_new
    else:  #this is a contacts (monopolar) objects
        elec_new = elec.copy()
        if 'atlases.mtl' in localization.keys():
            for atlas_idx, contact in enumerate(localization['atlases.mtl']['pairs'].index):
                try:
                    if np.logical_or(elec_new['label']==contact[0], elec_new['label']==contact[1]):
                        elec_new['mtl_atlas']=localization['atlases.mtl']['pairs'][atlas_idx]
                except:
                    continue
                
            return elec_new
        else:
            print('No MTL information found!')
            return elec_new

        
def updateElecs(localization, elecs):
    #Stolen from Ethan Solomon
    #For recent (R1380+) subjects who have MTL elecs in localization file but not pairs file

    #First check if bipolar or monopolar
    if 'contact_1' in elecs.keys(): #this is a pairs object
        pairs = elecs.copy()
        if 'atlases.mtl' in localization.keys():
            #Add the mtl atlas label to the original elecs structure
            for atlas_idx, pair in enumerate(localization['atlases.mtl']['pairs'].index):
                pairlabel = pair[0]+'-'+pair[1]
                try:
                    #Find where in the elecs structure the mtl atlas electrode is located
                    row_idx = np.where(pairs['label']==pairlabel)[0][0]
                    row_loc = pairs.index[row_idx]
                except:
                    continue
                pairs.loc[row_loc, 'mtl_atlas'] = localization['atlases.mtl']['pairs'][atlas_idx]
            return pairs
        else:
            print('No MTL information found!')
            return pairs
    else:  #this is a contacts (monopolar) objects
        contacts = elecs.copy()
        if 'atlases.mtl' in localization.keys():
            for atlas_idx, contact in enumerate(localization['atlases.mtl']['pairs'].index):
                try:
                    row_idx = np.where(np.logical_or(contacts['label']==contact[0], contacts['label']==contact[1]))[0][0]
                    row_loc = contacts.index[row_idx]
                except:
                    continue
                contacts.loc[row_loc, 'mtl_atlas']=localization['atlases.mtl']['pairs'][atlas_idx]
            return contacts
        else:
            print('No MTL information found!')
            return contacts
        

def getGoodElecs(reader):
    #Returns only the elecs that are in any of the pairs

    elecs = reader.load('contacts')
    pairs = reader.load('pairs')
    
    labels_elecs = np.array(elecs['label'])
    labels_pairs = np.array(pairs['label'])
    elecs_filt = []
    for le in labels_elecs:
        neg, plu = "-"+le, le+"-"
        x = (any([neg in lp for lp in labels_pairs]) | any([plu in lp for lp in labels_pairs]))
        elecs_filt.append(x)
    elecs = elecs[elecs_filt]
    
    return elecs


def getSOZ(reader):
    #Use clinical info spreadsheet to get names of electrodes in ictal onset
    clin = pd.read_csv('/home1/daniel.rubinstein/Downloads/Liliana-Clinical-Info.csv')

    #Return the soz brain regions based on the electrodes in soz category
    soz = []
    try:
        elecs = reader.load('contacts')
    except:
        elecs = reader.load('pairs')
    
    try:
        elecs_cat = reader.load('electrode_categories')
    except:
        try:
            reader1 = CMLReader(reader.subject, reader.experiment, montage=0, localization=reader.localization)
            elecs_cat = reader1.load('electrode_categories')
            try:
                elecs = reader1.load('contacts')
                reader = reader1
            except:
                elecs = reader1.load('pairs')
                reader = reader1
        except Exception as e:
            #print(e.__class__.__name__, str(e))
            elecs_cat = {'soz':[]}
            print('No electrode categories file found!')
            
    try:
        subj_num = reader.subject[2:5].lstrip('0')
        clin_labels = clin[clin['Subject Number']==subj_num].iloc()[1]['SOZ'].split()
    except:
        clin_labels = []
        print('No info in clinical spreadsheet')
        
    for e in range(len(elecs)):
        elec = elecs.iloc()[e]
        
        #In case pairs are loaded instead of contacts
        elec_labels = elec['label'].split('-')
        #try to account for dashes in name
        if (len(elec_labels) > 2) & (len(elec_labels) % 2 == 0):
            n2 = int(len(elec_labels)/2)
            elec_labels = ['-'.join(elec_labels[0:n2]), '-'.join(elec_labels[n2:])]
        for label in elec_labels:
            #make everything lowercase
            label = elec['label'].lower()
            hem = getHem(elecs.iloc()[e:e+1])[0]
            reg = 'brain'
            cat_soz = [l.lower().strip() for l in elecs_cat['soz']]
            cat_soz_clin = [l.lower() for l in clin_labels]

            if (label in cat_soz) | (label in cat_soz_clin):
                cols = ['ind.region','tal.region','mni.region','avg.region']
                c = 0
                while (reg == 'brain') & (c<len(cols)):
                    if cols[c] in elec.keys():
                        if (elec[cols[c]] not in [None,'nan','none']) & (type(elec[cols[c]]) is not float):
                            reg = elec[cols[c]]
                    c+=1

                if reg == 'brain':
                    reg = getPriorityMTLrois(reader,elecs.iloc()[e:e+1])[0]
                    
                soz.append(hem+'_'+str(reg))
        
    
                
    #templobe = ['superiortemporal','middletemporal','inferiortemporal',
    #            'bankssts','fusiform','transversetemporal','entorhinal','temporalpole','parahippocampal']
    
    
    return np.unique(soz)


def getHem(elecs):
    hems = []
    coordcolumns_poss = ['mni.x','tal.x','avg.x','das.x','stein.x','ind.x','vox.x','wb.x']
    coordcolumns = [c for c in coordcolumns_poss if c in elecs.keys()]

    for e in range(len(elecs)):
        elec = elecs.iloc()[e]
        hem = 'b'
        for cc in coordcolumns:
            if elec[cc]!='NaN':
                if elec[cc] < 0:
                    hem = 'l'
                elif elec[cc] > 0:
                    hem = 'r'
                if elec[cc]!=0:
                    break
                
        hems.append(hem)
    
    return hems

def getCoordVars(elec,loc='ind'):
    cstr = loc+'.corrected.x'
    if loc+'.corrected.x' in elec:
        xcoord, ycoord, zcoord = loc+'.corrected.x',loc+'.corrected.y',loc+'.corrected.z'
    elif loc+'.dural.x' in elec:
        xcoord, ycoord, zcoord = loc+'.dural.x',loc+'.dural.y',loc+'.dural.z'
        
    return xcoord, ycoord, zcoord

def getCoords(elecs,loc='ind'):
    coords = []
    for e in range(len(elecs)):
        elec = elecs.iloc()[e]
        xcoord,ycoord,zcoord = getCoordVars(elec,loc)
        coord = np.array(elec[[xcoord, ycoord, zcoord]])
        coords.append(coord)
    
    return np.array(coords)


def getPriorityMTLroi(reader, elec, checkbad=True):
    #Function to divide electrode into either PHG or the Hippocampus
    
    #Initialize to 'none'
    roi = 'none'
    
    mtllist_dk = ['parahippocampal','entorhinal','perirhinal']
    ltclist = ['superiortemporal','middletemporal','inferiortemporal','fusiform','bankssts','temporalpole','transversetemporal']
    
    itglist = ['inferiortemporal']
    mtglist = ['middletemporal']
    stglist = ['superiortemporal','bankssts']

    hipplist_stein = ['Left CA1','Left CA2','Left CA3','Left DG','Left Sub','Right CA1','Right CA2','Right CA3','Right DG','Right Sub']
    mtllist_stein = ['Left EC','Left PHC','Left PRC','Right EC','Right PHC','Right PRC']
    
    hipplist_das = ['CA1','CA2','CA3','DG','SUB']
    mtllist_das = ['BA35','BA36','ERC','PHC']
    
    hipplist_mtl_atlas = ['CA1','DG','SUB']
    mtllist_mtl_atlas = ['BA35','BA36','ERC','PHC']
    
    hipplist_all = ['CA1','CA2','CA3','DG','Sub','SUB']
    hipplist_all_lower = ['ca1','ca2','ca3','dg','sub']
    
    mtllist_all = ['BA35','BA36','EC','ERC','PHC','PRC']
    mtllist_all_lower = ['ba35','ba36','ec','erc','phc','prc']
    
    WMlist_stein = ['Left Fusiform Gyrus WM','Left MTL WM','Right MTL WM']
    WMlist_stein_lower = ['left fusiform gyrus wm','left mtl wm','right mtl wm']
    
    check = True
    if checkbad:
        check = checkBadElec(reader, elec)
    
    #Update the elecs structure to include the new mtl atlas as field "mtl_atlas"
    try:
        localization = reader.load('localization')
        elec = updateElec(localization, elec)
    except:
        pass
    
    #Initialize the different atlas regions
    steinreg = 'none'
    dasreg = 'none'
    mtlatlasreg = 'none'
    indreg = 'none'

    #First check if the electrode is not bad
    if check:

        #Check if different roi columns exists
        if 'stein.region' in elec.keys():
            steinreg = str(elec['stein.region'])
            steinreg = steinreg.strip('"').lower()

        if 'das.region' in elec.keys(): 
            dasreg = str(elec['das.region'])
            dasreg = dasreg.strip('"').lower()

        if 'mtl_atlas' in elec.keys():
            mtlatlasreg = str(elec['mtl_atlas'])
            mtlatlasreg = mtlatlasreg.strip('"').lower()

        if 'ind.region' in elec.keys():
            indreg = str(elec['ind.region'])
            indreg = indreg.strip('"').lower()

        # Do localization without prioritizing one method over the other:
        steinhipp = any([r in steinreg for r in hipplist_all_lower])
        steinphg = any([r in steinreg for r in mtllist_all_lower])
        dashipp = any([r in dasreg for r in hipplist_all_lower])
        dasphg = any([r in dasreg for r in mtllist_all_lower])
        mtlatlashipp = any([r in mtlatlasreg for r in hipplist_all_lower])
        mtlatlasphg = any([r in mtlatlasreg for r in mtllist_all_lower])
        indphg = any([r in indreg for r in mtllist_dk])
        indltc = any([r in indreg for r in ltclist])
        inditg = any([r in indreg for r in itglist])
        indmtg = any([r in indreg for r in mtglist])
        indstg = any([r in indreg for r in stglist])

        if steinhipp | dashipp | mtlatlashipp:
            roi = 'hipp'
        elif steinphg | dasphg | mtlatlasphg | indphg:
            roi = 'phg'
        elif inditg:
            roi = 'itg'
        elif indmtg:
            roi = 'mtg'
        elif indstg:
            roi = 'stg'
        elif indltc:
            roi = 'ltc'

    return roi


def getRegions(reader, pairs):
    from localization import getCoordVars, getPriorityMTLrois
    # Function to get the pairs from frontal, temporal, and parietal lobes based on ind.region
    frontalRegions = ['superiorfrontal','caudalmiddlefrontal','rostralmiddlefrontal','parsopercularis','parsorbitalis','parstriangularis','lateralorbitofrontal','medialorbitofrontal','paracentral','precentral','frontalpole']
    temporalRegions = ['superiortemporal','middletemporal','inferiortemporal','bankssts','fusiform','transversetemporal','entorhinal','temporalpole','parahippocampal']
    parietalRegions = ['superiorparietal','inferiorparietal','supramarginal','postcentral','precuneus']
    
    mtlRegions = getPriorityMTLrois(reader, pairs, checkbad=False)
    
    idx_frontal = []
    idx_temporal = []
    idx_parietal = []
    i=-1
    for p, pair in pairs.iterrows():
        i=i+1
        xc, yc, zc = getCoordVars(pair)
        
        regionStr = 'ind.corrected.region'
        if 'ind.corrected.region' not in pair:
            if 'ind.dural.region' in pair:
                regionStr = 'ind.dural.region'
            else:
                regionStr = 'ind.region'
        
        if (pair[regionStr] in frontalRegions):
            idx_frontal.append(np.where(pairs.index==p)[0][0])

        # Get the MTL region
        if (pair[regionStr] in temporalRegions) | (mtlRegions[i] in ['phg','hipp','ltc','itg','mtg','stg']):
            idx_temporal.append(np.where(pairs.index==p)[0][0])

        if (pair[regionStr] in parietalRegions):
            idx_parietal.append(np.where(pairs.index==p)[0][0])


    pairs_frontal = pairs.iloc()[idx_frontal]
    pairs_temporal = pairs.iloc()[idx_temporal]
    pairs_parietal = pairs.iloc()[idx_parietal]
    
    return pairs_frontal, pairs_temporal, pairs_parietal


def getSubjWithReg(subjects,task,reglist):
    df = get_data_index("r1")

    # Return the subjects subset with coverage in a particular region (by dk atlas - ind.region)
    subsub = []
    failed = []
    for s, subj in enumerate(subjects):
        try:
            sessions = list(df[(df['subject']==subj) & (df['experiment']==task)]['session'])
            monts = list(df[(df['subject']==subj) & (df['experiment']==task)]['montage'])
            locs = list(df[(df['subject']==subj) & (df['experiment']==task)]['localization'])
            reader = CMLReader(subj, task, sessions[0], montage=monts[0], localization=locs[0]) #just look at session 0
            pairs = reader.load('pairs')

            regionStr = 'ind.corrected.region'
            if 'ind.corrected.region' not in pairs.columns:
                if 'ind.dural.region' in pairs.columns:
                    regionStr = 'ind.dural.region'
                else:
                    regionStr = 'ind.region'

            if np.any([r in np.array(pairs[regionStr].unique()) for r in reglist]):
                subsub.append(subj)

        except:
            failed.append(subj)
            print('Localization failed for',subj)
            
    return subsub, failed


############
# Module from John Sakon:
def Loc2PairsTranslation(pairs,localizations):
    # localizations is all the possible contacts and bipolar pairs locations
    # pairs is the actual bipolar pairs recorded (plugged in to a certain montage of the localization)
    # this finds the indices that translate the localization pairs to the pairs/tal_struct

    loc_pairs = localizations.type.pairs
    loc_pairs = np.array(loc_pairs.index)
    split_pairs = [pair.upper().split('-') for pair in pairs.label] # pairs.json is usually upper anyway but things like "micro" are not
    pairs_to_loc_idxs = []
    for loc_pair in loc_pairs:
        loc_pair = [loc.upper() for loc in loc_pair] # pairs.json is always capitalized so capitalize location.pairs to match (e.g. Li was changed to an LI)
        loc_pair = list(loc_pair)
        idx = (np.where([loc_pair==split_pair for split_pair in split_pairs])[0])
        if len(idx) == 0:
            loc_pair.reverse() # check for the reverse since sometimes the electrodes are listed the other way
            idx = (np.where([loc_pair==split_pair for split_pair in split_pairs])[0])
            if len(idx) == 0:
                idx = ' '
        pairs_to_loc_idxs.extend(idx)

    return pairs_to_loc_idxs # these numbers you see are the index in PAIRS frame that the localization.pairs region will get put

def get_elec_regions(localizations,pairs):
    from copy import copy
    # 2020-08-13 new version after consulting with Paul 
    # suggested order to use regions is: stein->das->MTL->wb->mni
    
    # 2020-08-26 previous version input tal_struct (pairs.json as a recArray). Now input pairs.json and localizations.json like this:
    # pairs = reader.load('pairs')
    # localizations = reader.load('localization')
    
    regs = []    
    atlas_type = []
    pair_number = []
    has_stein_das = 0
    
    # if localization.json exists get the names from each atlas
    if len(localizations) > 1: 
        # pairs that were recorded and possible pairs from the localization are typically not the same.
        # so need to translate the localization region names to the pairs...which I think is easiest to just do here

        # get an index for every pair in pairs
        loc_translation = Loc2PairsTranslation(pairs,localizations)
        loc_dk_names = ['' for _ in range(len(pairs))]
        loc_MTL_names = copy(loc_dk_names) 
        loc_wb_names = copy(loc_dk_names)
        for i,loc in enumerate(loc_translation):
            if loc != ' ': # set it to this when there was no localization.pairs
                if 'atlases.mtl' in localizations: # a few (like 5) of the localization.pairs don't have the MTL atlas
                    loc_MTL_names[loc] = localizations['atlases.mtl']['pairs'][i] # MTL field from pairs in localization.json
                    has_MTL = 1
                else:
                    has_MTL = 0 # so can skip in below
                loc_dk_names[loc] = localizations['atlases.dk']['pairs'][i]
                loc_wb_names[loc] = localizations['atlases.whole_brain']['pairs'][i]   
    for pair_ct in range(len(pairs)):
        try:
            pair_number.append(pair_ct) # just to keep track of what pair this was in subject
            pair_atlases = pairs.iloc[pair_ct] #tal_struct[pair_ct].atlases
            if 'stein.region' in pair_atlases: # if 'stein' in pair_atlases.dtype.names:
                test_region = str(pair_atlases['stein.region'])
                if (test_region is not None) and (len(test_region)>1) and \
                   (test_region not in 'None') and (test_region != 'nan'):
                    regs.append(test_region.lower())
#             if 'stein' in pair_atlases.dtype.names:  ### OLD WAY FROM TAL_STRUCT...leaving as example
#                 if (pair_atlases['stein']['region'] is not None) and (len(pair_atlases['stein']['region'])>1) and \
#                    (pair_atlases['stein']['region'] not in 'None') and (pair_atlases['stein']['region'] != 'nan'):
#                     regs.append(pair_atlases['stein']['region'].lower())
                    atlas_type.append('stein')
                    has_stein_das = 1 # temporary thing just to see where stein/das stopped annotating
                    continue # back to top of for loop
                else:
                    pass # keep going in loop
            if 'das.region' in pair_atlases:
                test_region = str(pair_atlases['das.region'])
                if (test_region is not None) and (len(test_region)>1) and \
                   (test_region not in 'None') and (test_region != 'nan'):
                    regs.append(test_region.lower())
                    atlas_type.append('das')
                    has_stein_das = 1
                    continue
                else:
                    pass
            if len(localizations) > 1 and has_MTL==1:             # 'MTL' from localization.json
                if loc_MTL_names[pair_ct] != '' and loc_MTL_names[pair_ct] != ' ':
                    if str(loc_MTL_names[pair_ct]) != 'nan': # looking for "MTL" field in localizations.json
                        regs.append(loc_MTL_names[pair_ct].lower())
                        atlas_type.append('MTL_localization')
                        continue
                    else:
                        pass
                else:
                    pass
            if len(localizations) > 1:             # 'whole_brain' from localization.json
                if loc_wb_names[pair_ct] != '' and loc_wb_names[pair_ct] != ' ':
                    if str(loc_wb_names[pair_ct]) != 'nan': # looking for "MTL" field in localizations.json
                        regs.append(loc_wb_names[pair_ct].lower())
                        atlas_type.append('wb_localization')
                        continue
                    else:
                        pass
                else:
                    pass
            if 'wb.region' in pair_atlases:
                test_region = str(pair_atlases['wb.region'])
                if (test_region is not None) and (len(test_region)>1) and \
                   (test_region not in 'None') and (test_region != 'nan'):
                    regs.append(test_region.lower())
                    atlas_type.append('wb')
                    continue
                else:
                    pass
            if len(localizations) > 1:             # 'dk' from localization.json
                if loc_dk_names[pair_ct] != '' and loc_dk_names[pair_ct] != ' ':
                    if str(loc_dk_names[pair_ct]) != 'nan': # looking for "dk" field in localizations.json
                        regs.append(loc_dk_names[pair_ct].lower())
                        atlas_type.append('dk_localization')
                        continue
                    else:
                        pass
                else:
                    pass
            if 'dk.region' in pair_atlases:
                test_region = str(pair_atlases['dk.region'])
                if (test_region is not None) and (len(test_region)>1) and \
                   (test_region not in 'None') and (test_region != 'nan'):
                    regs.append(test_region.lower())
                    atlas_type.append('dk')
                    continue
                else:
                    pass
            if 'ind.corrected.region' in pair_atlases: # I don't think this ever has a region label but just in case
                test_region = str(pair_atlases['ind.corrected.region'])
                if (test_region is not None) and (len(test_region)>1) and \
                   (test_region not in 'None') and (test_region not in 'nan'):
                    regs.append(test_region.lower())
                    atlas_type.append('ind.corrected')
                    continue
                else:
                    pass  
            if 'ind.region' in pair_atlases:
                test_region = str(pair_atlases['ind.region'])
                if (test_region is not None) and (len(test_region)>1) and \
                   (test_region not in 'None') and (test_region != 'nan'):
                    regs.append(test_region.lower())
                    atlas_type.append('ind')
                    # [tal_struct[i].atlases.ind.region for i in range(len(tal_struct))] # if you want to see ind atlases for comparison to above
                    # have to run this first though to work in ipdb: globals().update(locals())                  
                    continue
                else:
                    regs.append('No atlas')
                    atlas_type.append('No atlas')
            else: 
                regs.append('No atlas')
                atlas_type.append('No atlas')
        except AttributeError:
            regs.append('error')
            atlas_type.append('error')
    return np.array(regs),np.array(atlas_type),np.array(pair_number),has_stein_das
###########